//
// Copyright (C) 2003 Columbia Genome Center
// All Rights Reserved.
//
// Matrix.cc -- Create a matrix.
// vi: set ts=4
//
// $Id: CMatrix.cpp,v 1.5 2013/03/19 04:37:25 manju Exp $
//

#include <algorithm>
#include <fstream>
#include <iostream>
#include <list>

#ifdef __BCPLUSPLUS__   // For Borland Compiler
  #include <stlport/hash_map>
using namespace std;
#else
  #ifdef __GNUC__        // For GNU gcc compiler
    #if __GNUC__ < 3
      #include <hash_map.h>
using namespace std;
#else
      #include <ext/hash_map>
      #if __GNUC_MINOR__ == 0
using namespace std; // GCC 3.0
#else
using namespace __gnu_cxx; // GCC 3.1 and later
#endif
    #endif
  #endif
#endif

#include <set>
#include <sstream>
#include <iomanip>
#include <string>
#include <utility>
#include <vector>
#include <cmath>
#include <ctime>
#include "CMatrix.h"

using namespace std;
typedef unsigned int uint;

void
CMatrix::computeCondMatrix(Microarray_Set & data, Microarray_Set & candidate,
                           Mutual_Info & mi, Mutual_Info & cmi, int hubId,
                           double threshold, double eps, double percent,
                           map <int,int> & subnet)
{
    
    /* compute pair-wise MI between the TF and all targets */
    this->push_back( hash_map < int, Node > () );

    for (int i=0; i<data.Get_Num_Markers(); i++)
    {
      Node node;
      double mutual = calculateMI(data, mi, hubId, i);
      node.Set_Mutualinfo( mutual );
      node.Set_Connection(true);
      (* this)[0][i] = node;
    }

    if (mi.Get_Type() != Mutual_Info::CMI_AP) {
      /* compute delta MI for all (modulator, target) pair */
      vector < int > low;
      vector < int > high;
      for ( int i = 0; i < candidate.Get_Num_Markers(); i++ )
      {
        cout << "Processing candidate: " << candidate.Get_Marker_AffyId(i) << endl;
        candidate.getHighLowPercent( percent, i, low, high );
        // get the index of the candidate probe in the dataset; if it does not
        // exist in the dataset, -1 will be returned
        string probe = candidate.Get_Marker_AffyId(i);
        int controlId = data.Get_Marker_Id(probe);

        this->push_back( hash_map < int, Node > () );
        generateCondSubnet(2*i+1, data, cmi, hubId, controlId, low, threshold, eps, subnet);

        this->push_back( hash_map < int, Node > () );
        generateCondSubnet(2*(i+1), data, cmi, hubId, controlId, high, threshold, eps, subnet);

      }
    } else {
      vector < int > all;
      vector < int > none;
      /* compute conditional MI for all (modulator, target) pair */
      for ( int i = 0; i < candidate.Get_Num_Markers(); i++ )
      {
        cout << "Processing candidate: " << candidate.Get_Marker_AffyId(i) << endl;
        candidate.getHighLowPercent( 1.0, i, all, none );
        string probe = candidate.Get_Marker_AffyId(i);
        int controlId = data.Get_Marker_Id(probe);

        this->push_back( hash_map < int, Node > () );
        generateCondSubnetCMI(i+1, data, mi, hubId, controlId, all, threshold, eps, subnet);
      }
   }
}

void
CMatrix::generateCondSubnet(int idx, Microarray_Set & data, Mutual_Info & mutual,
                            int hubId, int controlId, vector< int > & v, double threshold, double eps, map <int, int> & subnet)
{
    for (int i=0; i<data.Get_Num_Markers(); i++)
    {
        Node node;
        if ( not( isSameGene(data,i,controlId) || isSameGene(data,hubId,controlId) ) )
        {
            // whether hubId and i represent the same gene will be checked by the function calculateMI()
            double mi = calculateMI(data, mutual, hubId, i, v);
            node.Set_Mutualinfo( mi );
            if ( mi >= threshold )
            {
                node.Set_Connection( true );
            }
        }
        (* this)[idx][i] = node;
    }

    if ( eps < 1 )
    {
        reduce(idx, data, mutual, v, threshold, eps, subnet);
    }
}

void 
CMatrix::generateCondSubnetCMI( int idx, Microarray_Set & data, Mutual_Info & mutual,
                             int hubId, int controlId, vector< int > & v,
                             double threshold, double eps, map <int,int> & subnet)
{
    for (int i=0; i<data.Get_Num_Markers(); i++)
    {
        Node node;
        if ( not( isSameGene(data,i,controlId) || isSameGene(data,hubId,controlId) ) )
        {
            // whether hubId and i represent the same gene will be checked by the function calculateMI()
            double cmi = calculateCMI(data, mutual, hubId, i, controlId);
            node.Set_Mutualinfo( cmi );
            node.Set_Connection( true );
        }
        (* this)[idx][i] = node;
    }

    if ( eps < 1 )
    {
        reduce(idx, data, mutual, v, threshold, eps, subnet);
    }
}

void
CMatrix::reduce( int row_idx, Microarray_Set & data, Mutual_Info & mutual,
                 vector < int > & v, double threshold, double epsilon, map <int, int> & subnet)
{
    SortDecreasing_ArrayValuePair sorter;
    hash_map < int, Node > & m = (* this)[row_idx];
    vector < ArrayValuePair > miVector;
    // put all other nodes that are connected to the current node on a vector for sorting
    for ( hash_map < int, Node >::const_iterator it = m.begin(); it != m.end(); it++ )
    {
        if (it->second.Get_Connection())
        {
            miVector.push_back( ArrayValuePair( it->first, it->second.Get_Mutualinfo() ) );
        }
    }
    // sort the vector
    sort( miVector.begin(), miVector.end(), sorter );

    // for each value in the vector (these are genes directly connected to row_idx.
    // i.e. geneId1 is connected to i)
    for ( uint j = 0; j < miVector.size(); j++ )
    {
        // (geneId1, valueAB) are the (id, value) of the selected gene, starting with the largest mi
        int geneId1 = miVector[j].getId();
        double valueAB = miVector[j].getValue();

        // Set the limits
        double minMI = valueAB / ( 1.0 - epsilon );
        // Loop on all the genes that have an equal or higher mutual information
        // (these genes are also connected to i) bool connected = true;
        for ( uint k = 0; k < j; k++ )
        {
            int geneId2 = miVector[k].getId();
            double valueAC = miVector[k].getValue();
            // We know that valueAB is smaller than valueAC by default; then tolerance is always positive
            if ( valueAC <= minMI )
            {
                // we can exit from this loop because all future values are smaller
                break;
            }
            // We know that valueAB is smaller than valueAC by default.
            // Therefore, if it is also smaller than valueBC, then
            // geneId1 is connected to gene i via geneId2
            map <int, int>::iterator p1=subnet.find(geneId1);
            map <int, int>::iterator p2=subnet.find(geneId2);

            // break the loop only if at least two are TFs. The hub is TF for sure,
            // so only require one more to be in the subnet list
            // if ( subnet.empty() || (p1 != subnet.end()) || (p2 != subnet.end()) ) <--kai

            // wk: need to test only if the intermediate node (geneId2) is TF
            if ( subnet.empty() || (p2 != subnet.end()) )
            {
                double valueBC = calculateMI( data, mutual, geneId1, geneId2, v );
                if ( valueBC > minMI )
                {
                    Node & node = m[geneId1];
                    node.Set_Intermediate( geneId2 );
                    node.Set_Connection( false );
                    break;
                }
            }
        }
    }
}

/* Method for computing MI */
double
CMatrix::calculateMI( Microarray_Set & data, Mutual_Info & mutual,
                      int probeId1, int probeId2, vector < int > & v ) throw(string)
{
    if (mutual.Get_Num_Microarrays()!=v.size()) {
        cout << "mutual.Get_Num_Microarrays(): " << mutual.Get_Num_Microarrays() << endl;
        cout << "v.size(): " << v.size() << endl;
        throw(string("MI calculator does not match data!"));
    }

    double mi = 0;
    if (not(isSameGene(data,probeId1,probeId2)))
    {
        Pair_Vector pairs;
        for ( uint i = 0; i < v.size(); i++ )
        {
            Gene_Pair pair;
            Set_Pair( pair, data.Get_Value(v[i],probeId1), data.Get_Value(v[i],probeId2),i);
            pairs.push_back( pair );
        }
        mi = mutual.Compute_Pairwise_MI( pairs );
    }
    return mi;
}

/* Overloading methods for computing MI */
double
CMatrix::calculateMI( Microarray_Set & data, Mutual_Info & mutual,
                      int probeId1, int probeId2) throw(string)
{
    if (mutual.Get_Num_Microarrays()!=data.Get_Num_Microarrays()) {
        cout << "mutual.Get_Num_Microarrays(): " << mutual.Get_Num_Microarrays() << endl;
        cout << "data.Get_Num_Microarrays(): " << data.Get_Num_Microarrays() << endl;
            throw(string("MI calculator does not match data!"));
    }

    double mi = 0;
    if (not(isSameGene(data,probeId1,probeId2)))
    {
        Pair_Vector pairs;
        for ( uint i = 0; i < data.Get_Num_Microarrays(); i++ )
        {
            Gene_Pair pair;
            Set_Pair( pair, data.Get_Value(i,probeId1), data.Get_Value(i,probeId2),i);
            pairs.push_back( pair );
        }
        mi = mutual.Compute_Pairwise_MI( pairs );
    }
    return mi;
}

double 
CMatrix::calculateCMI( Microarray_Set & data, Mutual_Info & mutual,
                        int probeId1, int probeId2, int probeId3) throw(string)
{
    if (mutual.Get_Num_Microarrays()!=data.Get_Num_Microarrays()) {
        cout << "mutual.Get_Num_Microarrays(): " << mutual.Get_Num_Microarrays() << endl;
        cout << "data.Get_Num_Microarrays(): " << data.Get_Num_Microarrays() << endl;
            throw(string("MI calculator does not match data!"));
    }

    double cmi = 0;
    if (not(isSameGene(data,probeId1,probeId2)) && not(isSameGene(data,probeId1,probeId3)) && not(isSameGene(data,probeId2,probeId3)))
    {
        Triplet_Vector triplets;
        for ( uint i = 0; i < data.Get_Num_Microarrays(); i++ )
        {
            Gene_Triplet triplet;
            Set_Triplet( triplet, data.Get_Value(i,probeId1), data.Get_Value(i,probeId2), data.Get_Value(i,probeId3), i);
            triplets.push_back( triplet );
        }
        cmi = mutual.Get_Conditional_Mutual_Info_XYZ( triplets );
    }
    return cmi;
}

/* Overloading methods for computing MI */
double
CMatrix::calculateMI( Microarray_Set & data1, int probeId1,
                      Microarray_Set & data2, int probeId2,
                      Mutual_Info & mutual ) throw(string)
{
    if (mutual.Get_Num_Microarrays()!=data1.Get_Num_Microarrays())
        throw(string("MI calculator does not match data!"));

    if (data1.Get_Num_Microarrays()!=data2.Get_Num_Microarrays())
        throw(string("Vectors must be of same size!"));

    double mi = 0;
    if (not(isSameGene(data1, probeId1, data2, probeId2)))
    {
        Pair_Vector pairs;
        for ( uint i = 0; i < data1.Get_Num_Microarrays(); i++ )
        {
            Gene_Pair pair;
            Set_Pair( pair,data1.Get_Value(i,probeId1),data2.Get_Value(i,probeId2),i);
            pairs.push_back( pair );
        }
        mi = mutual.Compute_Pairwise_MI( pairs );
    }
    return mi;
}

bool
CMatrix::isSameGene( Microarray_Set & data, int probeId1, int probeId2 )
{
    if (probeId1==-1 || probeId2 == -1)
    {
        return false;
    }
    return ( (probeId1 == probeId2) ||
             ( ( data.Get_Marker_Label(probeId1) == data.Get_Marker_Label(probeId2) ) &&
               ( data.Get_Marker_Label(probeId1) != "---" ) )
           );
}


bool
CMatrix::isSameGene( Microarray_Set & data1, int probeId1,
                     Microarray_Set & data2, int probeId2 )
{
    if (probeId1==-1 || probeId2 == -1)
    {
        return false;
    }
    return ( (probeId1 == probeId2) ||
             ( ( data1.Get_Marker_Label(probeId1) == data2.Get_Marker_Label(probeId2) ) &&
               ( data1.Get_Marker_Label(probeId1) != "---" ) )
           );
}


inline double CMatrix::getNodeMI( int colid, int rowid )
{
    Node & node = (* this)[colid][rowid];
    return node.Get_Mutualinfo();
}

void CMatrix::write( Microarray_Set & data, Microarray_Set & candidate, int hubId,
                     string filename, bool reduced, int type )
{
    setWriteReduced( reduced );
    std::fstream output( filename.c_str(), std::ios::out );
    cout << "Writing matrix: " << filename << endl;

    vector <string> conditions = candidate.Get_Marker_Set();
    writeHeader( output, data, conditions, hubId, type );

    int num_conditions = conditions.size();
    for ( int i=0; i<data.Get_Num_Markers(); i++ )
    {
        writeLine( i, output, data, num_conditions, type );
    }
    output.flush();
    output.close();
}


void CMatrix::writeHeader( std::ostream & out, Microarray_Set & data, vector <string> & conditions,
                           int hubId, int type )
{
    out << "HUB::" << data.Get_Marker_AffyId(hubId) << "\tMI";

    if (type != Mutual_Info::CMI_AP) {
      for (uint i=0; i<conditions.size(); i++)
      {
          string m = conditions[i];
          out << "\t" << m << ":L\t" << m << ":H";
      }
      out << endl;
    } else {
      for (uint i=0; i<conditions.size(); i++)
      {
          string m = conditions[i];
          out << "\t" << m;
      }
      out << endl;
    }
}


void CMatrix::writeLine( int gid, ostream & out, Microarray_Set & data, int num_conditions, int type )
{
    Node & node = (* this)[0][gid];
    out << data.Get_Marker_AffyId(gid) << '\t' << node.Get_Mutualinfo();

    if (type != Mutual_Info::CMI_AP) {
      for (int i=0; i<num_conditions; i++)
      {
          Node & node1 = (* this)[2*i+1][gid];
          Node & node2 = (* this)[2*(i+1)][gid];

          if (writeReduced)
          {
              double value1 = node1.Get_Mutualinfo();
              double value2 = node2.Get_Mutualinfo();
              if (!node1.Get_Connection() && !node2.Get_Connection()){
                 value1 = 0;
                 value2 = 0;
	         out << "\t" <<  (int)value1 << "\t" << (int)value2;
              }
              else 
                 out << "\t" << setprecision(3) << value1 << "\t" << setprecision(3) << value2;
          }
          else
          {
              out << "\t" << setprecision(3) << node1.Get_Mutualinfo() << "\t" << setprecision(3) << node2.Get_Mutualinfo();
          }
      }
      out << endl;
    } else {
      for (int i=0; i<num_conditions; i++)
      {
          Node & node1 = (* this)[i+1][gid];
          out << "\t" << setprecision(5) << node1.Get_Mutualinfo();
      }
      out << endl;
    }
}


//
// reads an adjacency matrix from an input stream
//
void
CMatrix::readCMM(const string &file, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string)
{
    ifstream in(file.c_str());
    if(!in.good())
    {
        cerr << "Bad File" << endl;
        throw(std::string("Problem with reading the file."));
    }
    cout << "Reading CMM matrix from file ..." << endl;
    readCMM(in, data, hubId, conditions);
}


void
CMatrix::readCMM(std::istream& in, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string)
{
    if(!in.good())
    {
        throw(std::string("Bad stream."));
    }

    // read the header line, which contains all conditions in strings
    string header;
    getline(in, header);
    istringstream hin( header );

    string hub;
    getline(hin, hub, '\t');
    unsigned int pos = hub.find( "::" );
    hub = hub.substr( pos+2 ).c_str();
    hubId = data.Get_Marker_Id( hub );

    cout << "HUB: " << hub << endl;

    string low;
    string high;

    // by pass the unconditional MI column header
    string dummy;
    getline( hin, dummy, '\t');

    // first column is the unconditional MI
    this->push_back( hash_map < int, Node > () );

    int controlId = 0;
    while ( hin.good() )
    {
        getline( hin, low, '\t');
        getline( hin, high, '\t');

        low = low.substr(0, low.length()-2);
        high = high.substr(0, high.length()-2);

        if (low != high)
        {
            throw(string("Condition format wrong!"));
        }
        conditions.push_back(data.Get_Marker_Id(low));
        controlId++;
        this->push_back( hash_map < int, Node > () );
        this->push_back( hash_map < int, Node > () );
    }

    cout << "No. constraints: " << controlId << " (no. conditions: " << controlId*2 << ")" << endl;

    int gid = 0;
    string line;
    while ( in.good() && (in.peek() != EOF) )
    {
        getline(in, line);
        istringstream lin( line );

        string target;
        getline(lin, target, '\t');
        if (data.Get_Marker_Id( target ) != gid)
        {
            throw(string("Matrix format wrong!"));
        }

        double entry;
        int cid = 0;
        while ( lin.good() )
        {
            lin >> entry;
            Node node;
            node.Set_Mutualinfo( entry );
            (* this)[cid][gid] = node;
            cid++;
        }
        gid++;
    }
    cout << "No. genes: " << gid << endl;
}


void
CMatrix::readCAM(const string &file, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string)
{
    ifstream in(file.c_str());
    if(!in.good())
    {
        cout << "Corresponding CAM matrix does not exist." << endl;
        return;
    }
    cout << "Reading thresholding and DPI info. from the corresponding CAM matrix ..." << endl;
    readCAM(in, data, hubId, conditions);
}


void
CMatrix::readCAM(std::istream& in, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string)
{
    if(!in.good())
    {
        throw(std::string("Bad stream."));
    }

    // read the header line, which contains all conditions in strings
    string header;
    getline(in, header);
    istringstream hin( header );

    string hub;
    getline(hin, hub, '\t');
    unsigned int pos = hub.find( "::" );
    hub = hub.substr( pos+2 ).c_str();
    if (hubId != data.Get_Marker_Id( hub ))
    {
        throw(string("File format does not match!"));
    }

    // by pass the unconditional MI column header
    string dummy;
    getline( hin, dummy, '\t');

    string low;
    string high;

    int controlId = 0;
    while ( hin.good() )
    {
        getline( hin, low, '\t');
        getline( hin, high, '\t');

        low = low.substr(0, low.length()-2);
        high = high.substr(0, high.length()-2);

        if (low != high)
        {
            throw(string("Condition format wrong!"));
        }

        if (data.Get_Marker_Id(low) != conditions[controlId])
        {
            throw(string("File format does not match!"));
        }
        controlId++;
    }

    int gid = 0;
    string line;
    while ( in.good() && (in.peek() != EOF) )
    {
        getline(in, line);
        istringstream lin( line );

        string target;
        getline(lin, target, '\t');
        if (data.Get_Marker_Id( target ) != gid)
        {
            throw(string("Matrix format wrong!"));
        }

        double entry;
        int cid = 0;
        while ( lin.good() )
        {
            lin >> entry;
            Node & node = (* this)[cid][gid];
            if (entry != 0)
            {
                if ( entry != node.Get_Mutualinfo() )
                {
                    throw(string("File content does not match!"));
                }
                else
                {
                    node.Set_Connection(true);
                }
            }
            cid++;
        }
        gid++;
    }
}



