/*
 * Copyright (C) 2003  Columbia Genome Center
 * All Rights Reserved.
 *
 * Matrix.h -- Adjacency Matrix Data Structure
 *
 * $Id: CMatrix.h,v 1.2 2013/03/16 22:01:18 manju Exp $
 */

/******************************************************************************
* Class CMatrix
*
* This class represents the conditional adjacency matrix data structure that
* consists a list of genes and their connections to the hub under a series of
* conditions.
*
* One can think of an adjacency matrix as a two dimensional array: each entry
* in the array is called a node; a node[i,j] will present if there is a
* connection between gene i and the hub gene under condition j, absent
* otherwise.
*
* Inner Class Node - represent an edge between two genes
*
*******************************************************************************
*/


#ifndef MATRIX_H__
  #define MATRIX_H__

#include <istream>
  #include <string>
  #include <vector>
  #include <map>
  #include "Microarray_Set.h"
  #include "MutualInfo.h"

#ifdef __BCPLUSPLUS__   // For Borland Compiler
    #include <stlport/hash_map>
using namespace std;
#else
    #ifdef __GNUC__        // For GNU gcc compiler
      #if __GNUC__ < 3
        #include <hash_map.h>
using namespace std;
#else
        #include <ext/hash_map>
        #if __GNUC_MINOR__ == 0
using namespace std; // GCC 3.0
#else
using namespace __gnu_cxx; // GCC 3.1 and later
#endif
      #endif
    #endif
  #endif


/****************************************************************************
* Node (Inner) Class
*
* A node is one entry in the adjacency matrix that represents an edge between
* two genes.
* */
class Node
{
    double mutinfo;
    bool connected;
    int intermediate;

public:
    Node( )
    {
        connected = false;
        intermediate = -1;
        mutinfo = 0.0;
    }

    bool Is_Active() const
    {
        return ( intermediate < 0 );
    }

    void Set_Intermediate( int id )
    {
        intermediate = id;
    }

    int Get_Intermediate() const
    {
        return ( intermediate );
    }

    void Set_Connection( bool status )
    {
        connected = status;
    }

    bool Get_Connection() const
    {
        return ( connected );
    }

    void Set_Mutualinfo( double m )
    {
        mutinfo = m;
    }

    double Get_Mutualinfo() const
    {
        return ( mutinfo );
    }
};


/****************************************************************************
* Class Matrix
*
* Conceptually a square matrix storing the connections between a set of genes
*
* */
class CMatrix : public std::vector < hash_map < int, Node > >
{
    bool writeReduced;
    bool writeEmptyLines;

public:
    CMatrix()
    {
        writeReduced = false;     // if true, intermediate nodes will be written
        writeEmptyLines = false;  // if true, empty lines will be written
    }

    ~CMatrix()
    {}

    void Set_Pair( Gene_Pair & p, double xId, double yId, int maId )
    {
        p.Set_X( xId );
        p.Set_Y( yId );
        p.Set_MaID( maId );
    }

    void Set_Triplet( Gene_Triplet & t, double xId, double yId, double zId, int maId )
    {
        t.Set_X( xId );
        t.Set_Y( yId );
	t.Set_Z( zId );
        t.Set_MaID( maId );
    }

    void setWriteReduced( bool status )
    {
        writeReduced = status;
    }

    void setWriteEmptyLines( bool status )
    {
        writeEmptyLines = status;
    }

    void computeCondMatrix( Microarray_Set & data, Microarray_Set & data1,
                            Mutual_Info & mi, Mutual_Info & cmi,
                            int hubId, double threshold, double eps, double percent,
                            map <int,int> & subnet);

    void generateCondSubnet( int idx, Microarray_Set & data, Mutual_Info & mutual,
                             int hubId, int controlId, vector< int > & v,
                             double threshold, double eps, map <int,int> & subnet);

    void generateCondSubnetCMI( int idx, Microarray_Set & data, Mutual_Info & mutual,
                             int hubId, int controlId, vector< int > & v,
                             double threshold, double eps, map <int,int> & subnet);

    double calculateMI( Microarray_Set & data, Mutual_Info & mutual,
                        int probeId1, int probeId2, vector < int > & v ) throw(string);

    double calculateMI( Microarray_Set & data, Mutual_Info & mutual,
                        int probeId1, int probeId2) throw(string);
    double calculateMI( Microarray_Set & data1, int probeId1,
                        Microarray_Set & data2, int probeId2,
                        Mutual_Info & mutual ) throw(string);

    double calculateCMI( Microarray_Set & data, Mutual_Info & mutual,
                        int probeId1, int probeId2, int probeId3) throw(string);

    bool isSameGene( Microarray_Set & data, int probeId1, int probeId2 );

    bool isSameGene( Microarray_Set & data1, int probeId1,
                     Microarray_Set & data2, int probeId2 );

    void reduce( int row_idx, Microarray_Set & data, Mutual_Info & mutual,
                 vector < int > & v, double threshold, double epsilon, map <int,int> & subnet);

    double getNodeMI( int colid, int rowid );

    bool protectedByTFLogic( map <int, int> & transfac, int geneId1, int geneId2, int geneId3 );

    void write( Microarray_Set & data, Microarray_Set & candidate, int hubId,
                string filename, bool reduced, int type );

    void writeHeader( ostream & out, Microarray_Set & data,
                      vector <string> & conditions, int hubId, int type );

    void writeLine( int idx, ostream & out, Microarray_Set & data, int num_conditions, int type );

    void readCMM(const string & file, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string);

    void readCMM(istream& in, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string);

    void readCAM(const string & file, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string);

    void readCAM(istream& in, Microarray_Set & data, int & hubId, vector <int> & conditions) throw(string);

};

#endif
