// Copyright (C) 2003  Columbia Genome Center
// All Rights Reserved.
//
// Microarray_Set.h -- Microarray Data structures.
//
// $Id: Microarray_Set.h,v 1.1 2009/07/15 01:50:09 manju Exp $
//


/******************************************************************************
* Class Microarray_Set
*
* A Microarray dataset consists of multiple arrays, each of which measures the
* expression of thousands of "Markers". Here we represent each of these
* measurement using a "Probe".
*
* Inner Class Marker - represent markers on each array
*
* Inner Class Probe - represent a measurement on a marker in an array
*
* Inner Class ArrayValuePair
*             SortDecreasing_ArrayValuePair
*             SortIncreasing_ArrayValuePair
*   - datastructure for sorting the arrays be the
*   the expression values of a particular marker
*
*******************************************************************************
*/


#ifndef DAT_H__
  #define DAT_H__

#include <istream>
  #include <string>
  #include <vector>
  #include <map>
  #include <list>

#ifdef __BCPLUSPLUS__   // For Borland Compiler
    #include <stlport/hash_map>
using namespace std;
#else
    #ifdef __GNUC__        // For GNU gcc compiler
      #if __GNUC__ < 3
        #include <hash_map.h>
using namespace std;
#else
        #include <ext/hash_map>
        #if __GNUC_MINOR__ == 0
using namespace std; // GCC 3.0
#else
using namespace __gnu_cxx; // GCC 3.1 and later
#endif
      #endif
    #endif
  #endif

extern std::string null_string;
using namespace std;
typedef unsigned int uint;

/***********************************************************
 * Class Marker
 *
 * a marker describes a gene in a detailed way, including
 * attributes not useful in other places.
 */
class Marker
{
    int idnum;
    std::string accnum;
    std::string label;
    bool isActive;
    bool isControl;

public:
    Marker( int id = 0, const std::string & a = null_string, const std::string & l = null_string, bool active = true,
            bool control = false ) : idnum( id ), accnum( a ), label( l ), isActive( active ), isControl( control )
    {}

    int Get_ID() const
    {
        return idnum;
    }

    const string & Get_Accnum() const
    {
        return accnum;
    }

    const string & Get_Label() const
    {
        return label;
    }

    void Set_ID( int id )
    {
        idnum = id;
    }

    void Set_Accnum( std::string & a )
    {
        accnum = a;
    }

    void Set_Label( std::string & l )
    {
        label = l;
    }

    bool hasAccession( std::string & a )
    {
        // return ( accnum.find( a ) != accnum.npos );
        return ( accnum.compare(a)==0 );
    }

    void Enable()
    {
        isActive = true;
    }

    void Disable()
    {
        isActive = false;
    }

    bool Enabled() const
    {
        return ( isActive );
    }

    void set_Control()
    {
        isControl = true;
    }

    void set_No_Control()
    {
        isControl = false;
    }

    bool Is_Control()
    {
        return ( isControl );
    }

    friend std::ostream & operator << ( std::ostream & out_file, const Marker & m )
    {
        out_file << "(";
        out_file << m.label << "; " << m.accnum << "; ";
        out_file << m.idnum << "; ";
        if ( m.isActive )
            out_file << "Active; ";
        else
            out_file << "Not Active; ";
        if ( m.isControl )
            out_file << "Control)";
        else
            out_file << "Not Control)";

        return ( out_file );
    }
};



/*
 * A Marker_Set is an array of Markers, indexed by the
 * ID numbers contained within the marker.
 */
typedef std::vector < Marker > Marker_Set;



/***********************************************************
 * Class Probe
 *
 * Microarrays are really silicon chips that contain probes
 * to detect genes.  Each probe is a tuple containing values
 * and pvalues, and there are n such tuples arranged in an
 * array, indexed by the same numbers which indexes Markers
 * in a set.  Since Markers describe genes, the correspondence
 * is natural.  Associated with each set of Microarrays is
 * also a Marker_Set object.  If the cardinality of the marker
 * set is m, then each Microarray will have m probes
 *  associated with it.
 */
class Probe
{
    double val, pval;

public:
    /** deliminator btw probes when they are written to a stream */
    static const char * DELIMINATOR;

    Probe( double v = 0.0, double pv = 0.0 )
    {
        Set_Value( v );
        Set_PValue( pv );
    }

    double Get_Value() const
    {
        return val;
    }

    double Get_PValue() const
    {
        return pval;
    }

    void Set_Value( double v )
    {
        val = v;
    }

    void Set_PValue( double pv )
    {
        pval = pv;
    }


};

ostream & operator << ( ostream & out, const Probe & p );


typedef vector < Probe > Microarray;
typedef vector < Microarray > Microarray_Vector;


/*************************************************************
 * Class Microarray_Set
 *
 * Note that a Microarray_Set contains one or more microarrays
 * *and* has associated with it a set of markers.
 */
class Microarray_Set
{
    Marker_Set markerset;
    Microarray_Vector uarrays;
    map < string, int > markerMap;
    vector < string > header;
public:
    static const char * HEADER_DELIMINATOR;

    Microarray_Set()
    {}

    uint Get_Num_Microarrays() const { return uarrays.size(); }

    int Get_Num_Markers() const { return markerset.size(); }

    // return the number of active markers
    int Get_Num_Active_Markers();

    void Set_Array_Header(unsigned int i, const string & hdr);

    string Get_Array_Header(int i) { return header[i]; }

    const vector < string > & Get_Header() const { return header; }

    /***********************************************************
    *                   Marker manipulation
    ************************************************************/
    void Set_Marker(unsigned int i, const Marker & m);

    // return the Marker object at index i
    const Marker & Get_Marker(unsigned int i)const throw(string);

    // return the Marker object at index i
    Marker & Get_Marker(unsigned int i) throw(string);

    // return the index of a Marker given its string label
    int Get_Marker_Id(string & label);

    // return the marker accession number, i.e. Affymetrix IDs, for Marker i
    const string Get_Marker_AffyId(unsigned int i) throw(string);

    // return the marker label, i.e. gene symbol, for Marker i
    const string Get_Marker_Label(unsigned int i) throw(string);

    vector < string > Get_Marker_Set();

    /***********************************************************
    *                   Array manipulation
    ************************************************************/
    void Set_Microarray(unsigned i, const Microarray & ma);

    // return the Microarray Object, i.e. a vector of Probe objects, for Array i
    const Microarray & Get_Microarray(unsigned int i) const throw(string);

    /***********************************************************
    *                   Probe manipulation
    ************************************************************/
    void Set_Probe(unsigned int i, unsigned int j, Probe p);

    // return the Probe object at Array i and Marker m
    const Probe & Get_Probe(unsigned int i, unsigned int m) const throw(string);

    void Set_Value(unsigned int i, unsigned int m, double v) throw(string);

   // return the value at Array i, Marker m
    const double Get_Value(unsigned int i, unsigned int m) const throw(string);

    void Set_PValue(unsigned int i, unsigned int m, double v) throw(string);

    // return the p-value at Array i, Marker m
    const double Get_PValue(unsigned int i, unsigned int m) const throw(string);

    vector < double > Get_Probe_Value_Vector(unsigned int m) throw(string);

    unsigned int readMarkerWithPvalue( istream & in, const unsigned arrno ) throw( string );

    unsigned int readMarkerNoPvalue( istream & in, const unsigned arrno ) throw( string );

    unsigned int readHeader( istream & in ) throw( string );

    void Set_ColHeader( unsigned int i, const string & hdr );

    /* Read a matrix from a file. */
    void read( const std::string & fileName ) throw( std::string );

    /* Read a matrix from an input stream */
    void read( std::istream & in ) throw( std::string );

    // Used to filter all genes that have a mean lower than
    // minMean and a variance lower than mean * ratio. This limits
    // further analysis only to genes that have a reasonable dynamical range
    int filter( std::vector < int > & ids, double minMean = 50.0, double minSigma = 20.0, int ctlid = -1 );

    int filter( double minMean = 50.0, double sigma = 20.0, int ctlid = -1 );

    void getHighLowPercent( double x, int mId, vector < int > & lower, vector < int > & upper );

    void getRandomSubsamples( double x, int mId, vector < int > & lower, vector < int > & upper );

    void bootStrap( vector < int > & boot, vector < int > * arrays );

    void shuffleGene(unsigned int idx);
};

std::ostream & operator << ( std::ostream & out, const Microarray_Set & p );


class ArrayValuePair
{
    int arrayId;
    double value;
public:
    ArrayValuePair()
    {
        ;
    }
    ArrayValuePair(int Id, double v)
    {
        set(Id, v)
        ;
    }
    void set(int iId, double v)
    {
        arrayId = iId;
        value = v;
    }
    int getId() const
    {
        return arrayId;
    }
    double getValue() const
    {
        return value;
    }
};

class SortIncreasing_ArrayValuePair : std::binary_function<ArrayValuePair, ArrayValuePair, bool>
{
public:
    bool operator() (const ArrayValuePair& a, const ArrayValuePair& b) const
    {
        return(a.getValue() < b.getValue());
    }
};

class SortDecreasing_ArrayValuePair : std::binary_function<ArrayValuePair, ArrayValuePair, bool>
{
public:
    bool operator() (const ArrayValuePair& a, const ArrayValuePair& b) const
    {
        return(a.getValue() > b.getValue());
    }
};


#endif
