//
// Copyright (C) 2003 Columbia Genome Center
// All Rights Reserved.
//
// MutualInfo.cpp -- Mutual Information calculations. Equivalent
// of Java GeneGeneRelationship.
// vi: set ts=4
//
// $Id: MutualInfo.cpp,v 1.3 2013/03/16 22:01:07 manju Exp $
//

#include <algorithm>
#include <list>
#include <map>
#include <string>
#include <vector>
#include <fstream>
#include <iostream>
#include <sstream>
#include <math.h>
#include <cmath>
#include <ctime>
#include <valarray>
#include "MutualInfo.h"

#define STEPS 50

using namespace std;
typedef unsigned int uint;

Mutual_Info::Mutual_Info( int maNo, int _miSteps, double sigma, int _type )
        : Microarray_Num( maNo ), type( _type ), variance2( 2*sigma*sigma )
{
    switch ( type )
    {
        case MI_AVERAGE:
        miSteps = _miSteps;
        MA_Per_MI_Step = ( double )Microarray_Num / ( double )miSteps;
        MI_Space = new int * [miSteps];
        MI_Prob = new double * [miSteps];
        for ( int i = 0; i < miSteps; i++ )
        {
            MI_Space[i] = new int[miSteps];
            MI_Prob[i] = new double[miSteps];
            for ( int j = 0; j < miSteps; j++ )
            {
                MI_Space[i] [j] = 0;
                MI_Prob[i] [j] = 0.0;
            }
        }
        break;

        case MI_GAUSSIAN:

        // set coupla transform
        Set_Copula_Transform( true );

        // initialize the lookup table for computing Gaussian kernels
        prob_table = new double * [maNo];
        for ( int i = 0; i < maNo; i++ )
        {
            prob_table[i] = new double[maNo];
            for ( int j = 0; j < maNo; j++ )
            {
                prob_table[i][j] = -1.0;
            }
        }

        // initialize the lookup table for computing normalization factors
        norm_1D_table = new double[maNo];

        norm_2D_table = new double * [maNo];
        for ( int p = 0; p < maNo; p++ )
        {
            norm_2D_table[p] = new double[maNo];
        }

        Initialize_Norm_Table(sigma);

        break;

        case MI_ADAPTIVE_PARTITIONING:
           Set_Copula_Transform( false );
        break;

        case CMI_AP:
           Set_Copula_Transform( false );
        break;

        case PEARSON:
        break;

        case SPEARMAN:
        break;

        default:
        throw std::string( "MI computation not supported" );
    }
}

void Mutual_Info::Initialize_Norm_Table( double sigma)
{
    // initialize the 1D table
    for ( uint i=0; i<Microarray_Num; i++ )
    {
        // array id is from 0 to M-1, after copula transform, it becomes from 0 to 1
        // with 1/(2M) at both sides
        double x = Copula_Transform ? ( ( double(i+1)-0.5 )/Microarray_Num ) : double( i );
        norm_1D_table[i] = 0.5*( erf( (1-x)/(sigma*M_SQRT2) ) - erf( (0-x)/(sigma*M_SQRT2) ) );
    }

    // initialize the 2D table
    for (uint j=0; j<Microarray_Num; j++)
    {
        for ( uint k=j; k<Microarray_Num; k++ )
        {
            norm_2D_table[j][k] = norm_1D_table[j]*norm_1D_table[k];
            norm_2D_table[k][j] = norm_2D_table[j][k];
        }
    }
}

/** Get Score for MI_GAUSSIAN */
double Mutual_Info::Compute_Pairwise_MI( Pair_Vector & pairs )
{
    double mi;
    switch ( type )
    {
        case Mutual_Info::MI_AVERAGE:
        mi = Get_Mutual_Info_XY( pairs, Loki::Int2Type < Mutual_Info::MI_AVERAGE > () );
        break;

        case Mutual_Info::MI_ADAPTIVE_PARTITIONING:
         mi = Get_Mutual_Info_XY( pairs, Loki::Int2Type < Mutual_Info::MI_ADAPTIVE_PARTITIONING > () );
        break;

        case Mutual_Info::MI_GAUSSIAN:
        mi = Get_Mutual_Info_XY( pairs, Loki::Int2Type < Mutual_Info::MI_GAUSSIAN > () );
        break;

        case Mutual_Info::PEARSON:
        mi = Get_Mutual_Info_XY( pairs, Loki::Int2Type < Mutual_Info::PEARSON > () );
        break;

        case Mutual_Info::SPEARMAN:
        mi = Get_Mutual_Info_XY( pairs, Loki::Int2Type < Mutual_Info::SPEARMAN > () );
        break;

        case Mutual_Info::CMI_AP:
        mi = Get_Mutual_Info_XY( pairs, Loki::Int2Type < Mutual_Info::MI_ADAPTIVE_PARTITIONING > () );
        break;

        default:
        throw std::string( "MI computation not supported" );
    }
    return ( mi );
}


Mutual_Info::~Mutual_Info()
{
    switch ( type )
    {
        case MI_AVERAGE:
        for ( int i = 0; i < miSteps; i++ )
        {
            delete[] MI_Space[i];
            delete[] MI_Prob[i];
        }
        delete[] MI_Space;
        delete[] MI_Prob;
        break;

        case MI_ADAPTIVE_PARTITIONING:
        break;

	case CMI_AP:
	break;

        case MI_GAUSSIAN:
        for ( uint i = 0; i < Microarray_Num; i++ )
        {
            delete[] prob_table[i];
            delete[] norm_2D_table[i];
        }

        delete[] prob_table;
        delete[] norm_2D_table;
        delete[] norm_1D_table;

        break;
    }
}


inline void Mutual_Info::randomize_X(Gene_Pair & gp)
{
    // add an infiniestimal amount
    double val = gp.Get_X()+0.00001*((static_cast<double>(rand())/RAND_MAX)-0.5);
    gp.Set_X(val);
}


inline void Mutual_Info::randomize_Y(Gene_Pair & gp)
{
    // add an infiniestimal amount
    double val = gp.Get_Y()+0.00001*((static_cast<double>(rand())/RAND_MAX)-0.5);
    gp.Set_Y(val);
}


double Mutual_Info::Get_Mutual_Info_XY( Pair_Vector & pairs, const Loki::Int2Type < Mutual_Info::MI_GAUSSIAN > & rankType )
{
    const uint size = pairs.size();

    // rank and copula transformation
    Sort_X X_Sorter;
    sort( pairs.begin(), pairs.end(), X_Sorter );
    // randomize ties
    for( uint i=1; i<size; i++)
    {
        Gene_Pair & gp = pairs[i-1];
        if(gp.Get_X() == pairs[i].Get_X())
        {
            randomize_X(gp);
        }
    }
    sort(pairs.begin(), pairs.end(), X_Sorter);
    for ( uint i = 0; i < size; i++ )
    {
        pairs[i].Set_XI(i);
        if (Copula_Transform)
        {
            pairs[i].Set_X( ( double(i+1)-0.5 )/size );
        }
    }

    Sort_Y Y_Sorter;
    sort( pairs.begin(), pairs.end(), Y_Sorter );
    // randomize ties
    for( uint i=1; i<size; i++)
    {
        Gene_Pair & gp = pairs[i-1];
        if(gp.Get_Y() == pairs[i].Get_Y())
        {
            randomize_Y(gp);
        }
    }
    sort( pairs.begin(), pairs.end(), Y_Sorter );
    for ( uint i = 0; i < size; i++ )
    {
        pairs[i].Set_YI( i );
        if (Copula_Transform)
        {
            pairs[i].Set_Y( ( double(i+1)-0.5 )/size );
        }
    }

    double sum = 0.0;
    for ( uint i=0; i<size; i++ )
    {
        double v = Get_Kernel( pairs, i );
        sum += log( v );
    }
    // double mi = sum / (static_cast <double>size);
    double mi = sum/(double)size;
    return ( std::max( mi, 0.0 ) );

}

double Mutual_Info::Get_Kernel( Pair_Vector & pairs, unsigned int i )
{
    // mu_x, mu_y - the center of each gaussian kernel
    // ix, iy     - distance to the gaussian center in unit of index.
    // dx, dy     - the actual distance used to compute the kernel denstiy.
    //              if compula transformed, they are equal to ix and iy rescaled
    //              between 0 and 1;

    double fxy = 0.0;
    double fx = 0.0;
    double fy = 0.0;

    const uint size = pairs.size();

    for ( uint j = 0; j < size; j++ )
    {
        int mu_x = pairs[j].Get_XI();
        int mu_y = pairs[j].Get_YI();

        int ix = std::abs( pairs[i].Get_XI() - mu_x );
        int iy = std::abs( pairs[i].Get_YI() - mu_y );

        double dx = pairs[i].Get_X() - pairs[j].Get_X();
        double dy = pairs[i].Get_Y() - pairs[j].Get_Y();

        // compute the kernel density as necessary
        if ( prob_table[ix][iy] == -1.0 )
        {
            // if (ix, iy) is not computed, at least one of ix and iy is not computed
            // otherwise, all three should have been computed
            // when either ix=0 or iy=0, the trctdGssnBi.getProbability(dx, dy) is
            // equal to trctdGssnUni.getProbability(dx) or trctdGssnUni.getProbability(dy)
            if ( prob_table[ix][0] == -1.0 )
            {
                prob_table[ix][0] = std::exp(-(dx*dx)/variance2);
                prob_table[0][ix] = prob_table[ix][0];
            }

            if ( prob_table[0][iy] == -1.0 )
            {
                prob_table[0][iy] = std::exp(-(dy*dy)/variance2);
                prob_table[iy][0] = prob_table[0][iy];
            }

            prob_table[ix][iy] = prob_table[ix][0]*prob_table[0][iy];
            prob_table[iy][ix] = prob_table[ix] [iy];
        }

        fx += prob_table[ix][0]/norm_1D_table[mu_x];
        fy += prob_table[0][iy]/norm_1D_table[mu_y];
        fxy += prob_table[ix][iy]/norm_2D_table[mu_x][mu_y];
    }

    return ( ( fxy * size ) / ( fx * fy ) );
}


double Mutual_Info::Get_Mutual_Info_XY( Pair_Vector & pairs, const Loki::Int2Type < Mutual_Info::MI_AVERAGE > & rankType )
{
    double H_xy = 0;
    double H_x = -log( 1 / ( double )miSteps );
    double H_y = H_x;
    Sort_X X_Sorter;
    Sort_Y Y_Sorter;

    sort( pairs.begin(), pairs.end(), X_Sorter );
    for ( uint j = 0; j < pairs.size(); j++ )
    {
        pairs[j].Set_XI( j );
    }
    sort( pairs.begin(), pairs.end(), Y_Sorter );
    for ( uint j = 0; j < pairs.size(); j++ )
    {
        pairs[j].Set_YI( j );
    }
    for ( int i = 0; i < miSteps; i++ )
    {
        for ( int j = 0; j < miSteps; j++ )
        {
            MI_Space[i] [j] = 0;
            ;
        }
    }
    for ( uint i = 0; i < pairs.size(); i++ )
    {
        int x = ( int )( ( double )pairs[i].Get_XI() / MA_Per_MI_Step );
        int y = ( int )( ( double )pairs[i].Get_YI() / MA_Per_MI_Step );
        MI_Space[x] [y] ++;
    }

    // Compute the value for the initial block
    int count = 0;
    for ( int di = 0; di < MIBLOCKS; di++ )
    {
        for ( int dj = 0; dj < MIBLOCKS; dj++ )
        {
            count += MI_Space[di] [dj];
        }
    }
    double sumP = 0;
    for ( int i = 0; i < miSteps; i++ )
    {
        int count1 = count;
        for ( int j = 0; j < miSteps; j++ )
        {
            double p = ( double )count1 / ( double )( pairs.size() * MIBLOCKS * MIBLOCKS );
            sumP += p;
            if ( p > 0 )
                H_xy -= p * log( p );
            MI_Prob[i] [j] = ( double )count1;
            for ( int off = 0; off < MIBLOCKS; off++ )
            {
                // Wrap around the torus
                int x = ( i + off ) % miSteps;
                int y = ( j + MIBLOCKS ) % miSteps;
                count1 -= MI_Space[x][j];
                count1 += MI_Space[x][y];
            }
        }
        for ( int off = 0; off < MIBLOCKS; off++ )
        {
            int x = ( i + MIBLOCKS ) % miSteps;
            count -= MI_Space[i][off];
            count += MI_Space[x][off];
        }
    }
    double mi = ( H_x + H_y - H_xy ) / ( H_x + H_y );
    return mi;
}

double Mutual_Info::Get_Mutual_Info_XY( Pair_Vector & pairs, const Loki::Int2Type < Mutual_Info::PEARSON > & rankType )
{
    const uint n = pairs.size();
    double sum_xy = 0;
    double sum_x = 0;
    double sum_y = 0;
    double sum_x2 = 0;
    double sum_y2 = 0;

    for ( uint i=0; i<n; i++)
    {
        sum_xy += pairs[i].Get_X()*pairs[i].Get_Y();
        sum_x += pairs[i].Get_X();
        sum_y += pairs[i].Get_Y();
        sum_x2 += pairs[i].Get_X()*pairs[i].Get_X();
        sum_y2 += pairs[i].Get_Y()*pairs[i].Get_Y();
    }

    return (sum_xy - sum_x*sum_y/n)/sqrt((sum_x2-sum_x*sum_x/n)*(sum_y2-sum_y*sum_y/n));
}

double Mutual_Info::Get_Mutual_Info_XY( Pair_Vector & pairs, const Loki::Int2Type < Mutual_Info::SPEARMAN > & rankType )
{
    const uint n = pairs.size();

    Sort_X X_Sorter;
    sort( pairs.begin(), pairs.end(), X_Sorter );
    // adjust for ties
    uint i = 0;
    double xadj = 0;
    while (i < n)
    {
        uint j = i+1;
        while (j < n)
        {
            if (pairs[j].Get_X() != pairs[i].Get_X()) {
                break;
            }
            j++;
        }
        // if there is no tie, j-1 is equal to i, then this equation becomes: pairs[i].Set_XI(i)
        for (uint k=i; k<j; k++) {
            // "+1" to get the rank
            pairs[k].Set_X(1+(i+j-1)/2);
        }

        uint ntied = j-i;
        xadj +=  ntied*(ntied-1)*(ntied+1)/2;

        i = j;
    }

    Sort_Y Y_Sorter;
    sort( pairs.begin(), pairs.end(), Y_Sorter );
    // adjust for ties
    i = 0;
    double yadj = 0;
    while (i < n)
    {
        uint j = i+1;
        while (j < n)
        {
            if (pairs[j].Get_Y() != pairs[i].Get_Y()) {
                break;
            }
            j++;
        }
        // if there is no tie, j-1 is equal to i, then this equation becomes: pairs[i].Set_XI(i)
        for (uint k=i; k<j; k++) {
            // "+1" to get the rank
            pairs[k].Set_Y(1+(i+j-1)/2);
        }

        uint ntied = j-i;
        yadj +=  ntied*(ntied-1)*(ntied+1)/2;

        i = j;
    }

    // Spearman rank correlation is equivalant to Pearson correlation on rank
    // transformed; but below is a simpler way compute this quantity
    double D = 0;
    for (i=0; i<n; i++)
    {
        D += (pairs[i].Get_X()-pairs[i].Get_Y())*(pairs[i].Get_X()-pairs[i].Get_Y());
    }
    uint n3const = (n+1)*n*(n-1)/3;
    double meanD = (n3const - (xadj+yadj)/3)/2;
    double stdD  = sqrt((n3const/2 - xadj/3)*(n3const/2 - yadj/3)/(n-1));

    return (meanD - D)/(sqrt((double)(n-1))*stdD);
}

double Mutual_Info::Get_Mutual_Info_XY( Pair_Vector & pairs, const Loki::Int2Type < Mutual_Info::MI_ADAPTIVE_PARTITIONING > & rankType )
{
  Sort_X X_Sorter;
  for ( uint i = 0; i < pairs.size(); i++ )
  {
    pairs.at(i).Set_XI(i);
    pairs.at(i).Set_YI(i);
  }
  sort( pairs.begin(), pairs.end(), X_Sorter );
  valarray<int> xranks(pairs.size());
  for (uint i = 0; i < pairs.size(); i++) {
    xranks[pairs.at(i).Get_XI()] = i + 1;
  }

  Sort_Y Y_Sorter;
  sort( pairs.begin(), pairs.end(), Y_Sorter );
  valarray<int> yranks(pairs.size());
  for (uint i = 0; i < pairs.size(); i++) {
    yranks[pairs.at(i).Get_YI()] = i + 1;
  }

  double xcor = 0; int npar = 1; int run = 0; int N = pairs.size();
  valarray<int> poc(1, 20); valarray<int> kon(N, 20); valarray<int> poradi(N);
  valarray<int> NN(4); valarray<int> marg(0, 80);

  for (int i = 1; i <= N; i++) {
    poradi[i-1] = i;
  }

  int t[] = {1, 1, N, N};
  marg[slice(0, 4, 20)] = valarray<int>(t, 4);

  while (npar > 0) {
    run++;
    int apoc = poc[npar - 1];
    int akon = kon[npar - 1];
    valarray<int> apor(akon - apoc + 1);
    for (int i = 0; i <= (akon - apoc); i++) {
      apor[i] = poradi[apoc + i - 1];
    }

    int Nex = apor.size();

    valarray<int> rownpar = marg[slice((npar - 1), 4, 20)];

    int ave1 = (int)floor((rownpar[0] + rownpar[2]) / 2);
    int ave2 = (int)floor((rownpar[1] + rownpar[3]) / 2);

    valarray<bool> J1(apor.size());

    valarray<size_t> apor1 = valarray<size_t>(apor.size());

    for (uint i = 0; i < apor.size(); i++) {
      apor1[i] = apor[i] - 1;
    }

    valarray<int> maskedxranks = valarray<int>(xranks[apor1]);

    J1[maskedxranks <= ave1] = true; J1[maskedxranks > ave1] = false;

    valarray<bool> J2(apor.size());

    valarray<int> maskedyranks = valarray<int>(yranks[apor1]);
    J2[maskedyranks <= ave2] = true; J2[maskedyranks > ave2] = false;

    valarray<bool> Ia(4 * apor.size());

    Ia[slice(0, apor.size(), 4)] = J1 && J2;
    Ia[slice(1, apor.size(), 4)] = J1 && (!J2);
    Ia[slice(2, apor.size(), 4)] = (!J1) && J2;
    Ia[slice(3, apor.size(), 4)] = (!J1) && (!J2);


    valarray<int> I( 4 * apor.size());

    I[Ia] = 1; I[!Ia] = 0;

    NN[0] = (valarray<int>(I[slice(0, apor.size(), 4)])).sum();
    NN[1] = (valarray<int>(I[slice(1, apor.size(), 4)])).sum();
    NN[2] = (valarray<int>(I[slice(2, apor.size(), 4)])).sum();
    NN[3] = (valarray<int>(I[slice(3, apor.size(), 4)])).sum();

    valarray<int> amarg(16);

    valarray<int> slicemarg = valarray<int>(marg[slice((npar - 1), 4, 20)]);

    int t[] = { slicemarg[0], slicemarg[1] , ave1, ave2 };
    amarg[slice(0, 4, 4)] = valarray<int>(t, 4);

    int t1[] = { slicemarg[0], ave2 + 1, ave1, slicemarg[3] };
    amarg[slice(1, 4, 4)] = valarray<int>(t1, 4);

    int t2[] = { ave1 + 1, slicemarg[1], slicemarg[2], ave2 };
    amarg[slice(2, 4, 4)] = valarray<int>(t2, 4);

    int t3[] = { ave1 + 1, ave2 + 1, slicemarg[2], slicemarg[3] };
    amarg[slice(3, 4, 4)] = valarray<int>(t3, 4);

    valarray<double> NN2(4);
    NN2[0] = (double)NN[0];
    NN2[1] = (double)NN[1];
    NN2[2] = (double)NN[2];
    NN2[3] = (double)NN[3];

    double tst = 4 * valarray<double>(pow(NN2 - ((double)Nex/4.0), 2)).sum() / Nex;

    if (tst > 7.8 || run == 1) {
      --npar;
      for (uint i = 0; i < 4; i++) {
        if (NN[i] > 2) {
          ++npar;
          akon = apoc + NN[i] - 1;
          poc[npar - 1] = apoc;
          kon[npar - 1] = akon;
          marg[slice((npar - 1), 4, 20)] = amarg[slice(i, 4, 4)];

          valarray<int> indices = valarray<int>(I[slice(i, apor.size(), 4)]);
          valarray<int> t = valarray<int>(apor[indices==1]);
          for (int kk = (apoc - 1); kk < akon; kk++){
            poradi[kk] = t[kk - apoc + 1];
          }
          apoc = akon + 1;
        }
        else if (NN[i] > 0) {
          valarray<int> t = valarray<int>(amarg[slice(i, 4, 4)]);
          double Nx = t[2] - t[0] + 1;
          double Ny = t[3] - t[1] + 1;
          xcor += NN[i] * log(NN[i] / (Nx * Ny));
        }
      }
    }
    else {
      valarray<int> t = valarray<int>(marg[slice((npar - 1), 4, 20)]);
      double Nx = t[2] - t[0] + 1;
      double Ny = t[3] - t[1] + 1;
      xcor += Nex * log(Nex / (Nx * Ny));
      --npar;
    }
  }
  xcor = xcor / N + log(N);
  return xcor;
}

template <typename T>
void print(ostream& out, const valarray<T>& theValArray)
{
 for (int i = 0; i < theValArray.size(); i++)
 out << theValArray[i] << " ";
}

/* reverse:  reverse string s in place */
 void reverse(char s[])
 {
     int i, j;
     char c;

     for (i = 0, j = strlen(s)-1; i<j; i++, j--) {
         c = s[i];
         s[i] = s[j];
         s[j] = c;
     }
 }

void itoa(int n, char s[], int length)
 {
     int i, sign;

     if ((sign = n) < 0)  /* record sign */
         n = -n;          /* make n positive */
     i = 0;
     do {       /* generate digits in reverse order */
         s[i++] = n % 2 + '0';   /* get next digit */
     } while ((n /= 2) > 0);     /* delete it */
     while (i < length) {
         s[i++] = '0';
     }
     if (sign < 0)
         s[i++] = '-';
     s[i] = '\0';
     reverse(s);
 }


double 
Mutual_Info::Get_Conditional_Mutual_Info_XYZ( Triplet_Vector & triplets) {
        double xcor = 0;
        int N = triplets.size();
        int dim = 3;

  Sort_XT X_Sorter;
  for ( uint i = 0; i < triplets.size(); i++ )
  {
    triplets.at(i).Set_XI(i);
    triplets.at(i).Set_YI(i);
    triplets.at(i).Set_ZI(i);
  }
  sort( triplets.begin(), triplets.end(), X_Sorter );
  valarray<int> xranks(triplets.size());
  for (uint i = 0; i < triplets.size(); i++) {
    xranks[triplets.at(i).Get_XI()] = i + 1;
  }

  Sort_YT Y_Sorter;
  sort( triplets.begin(), triplets.end(), Y_Sorter );
  valarray<int> yranks(triplets.size());
  for (uint i = 0; i < triplets.size(); i++) {
    yranks[triplets.at(i).Get_YI()] = i + 1;
  }

  Sort_ZT Z_Sorter;
  sort( triplets.begin(), triplets.end(), Z_Sorter );
  valarray<int> zranks(triplets.size());
  for (uint i = 0; i < triplets.size(); i++) {
    zranks[triplets.at(i).Get_ZI()] = i + 1;
  }
 
  valarray<int> ranks(dim * N);
  ranks[slice(0, N, dim)] = xranks;
  ranks[slice(1, N, dim)] = yranks;
  ranks[slice(2, N, dim)] = zranks;

  int ddim = (int)pow((double)2, dim);
  int dim2 = 2 * dim;
  int npar = 1;
  int run = 0;

  valarray<int> poc(1, 20); valarray<int> kon(N, 20); valarray<int> poradi(N);
  valarray<double> NN(ddim); valarray<int> marg((8 * ddim) * dim2);

  for (int i = 1; i <= N; i++) {
    poradi[i-1] = i;
  }

  valarray<int> t1(1, dim);
  valarray<int> t2(N, dim);
  valarray<int> t(dim2);
  t[slice(0, dim, 1)] = t1;
  t[slice(dim, dim, 1)] = t2;
  marg[slice(0, dim2, (8 * ddim))] = t;

  valarray<int> Imm(dim * ddim);
  for (int i = 0; i < ddim; i++) {
    valarray<int> val(dim);
    char* buffer = new char[dim];
    itoa(i, buffer, dim);
    for (int j = 0; j < dim; j++) {
      stringstream ss;
      string s;
      char c = buffer[j];
      ss << c;
      ss >> s;
      int v = atoi(s.c_str());
      val[j] = v; 
    }
    Imm[slice(i, dim, ddim)] = val;
    delete buffer;
  }

  valarray<double> chi2(5);
  chi2[0] = 0;
  chi2[1] = 7.81;
  chi2[2] = 13.9;
  chi2[3] = 25.0;
  chi2[4] = 42.0;

  while (npar > 0) {
    run++;
    int apoc = poc[npar - 1];
    int akon = kon[npar - 1];

    valarray<int> apor = poradi[slice(apoc -1, akon - apoc + 1, 1)];

    int Nex = apor.size();

    valarray<int> r1 = marg[slice(npar - 1, dim2, (8 * ddim))];
    valarray<int> r2 = r1[slice(0, dim, 1)];
    valarray<int> r3 = r1[slice(dim, dim2, 1)];
    valarray<int> ave = (r2 + r3)/2.0;

    valarray<bool> J(Nex * ave.size());
    valarray<int> temp(Nex * ave.size());
    valarray<int> temp1(Nex * ave.size());

    valarray<bool> indices(false, N);
    for (uint j = 0; j < apor.size(); j++) {
      int index = apor[j] - 1;
      indices[index] = true;
    }
    for (int j = 0; j < dim; j++) {
      valarray<int> r = ranks[slice(j, N, dim)];
      temp[slice(j, Nex, dim)] = r[indices==true];
      temp1[slice(j, Nex, ave.size())] = ave[j];
    }

    J = temp <= temp1;

    valarray<bool> I(Nex * ddim);

    valarray<int> amarg(ddim * dim2); 
    for (int j = 0; j < ddim; j++) {
      valarray<int> temp10 = marg[slice(npar - 1, dim2, (8 * ddim))];
      amarg[slice(j, dim2, ddim)] = temp10;
    }
    
    for (int d = 0; d < ddim; d++) {
      valarray<bool> temp2(true, Nex);
      I[slice(d, Nex, ddim)] = temp2;

      for (int k = 0; k < dim; k++) {
        valarray<int> temp3 = Imm[slice(d, dim, ddim)];
        if (temp3[k] == 1) {
	  valarray<bool> a = J[slice(k, Nex, ave.size())];
	  a = !a;
	  valarray<bool> b = I[slice(d, Nex, ddim)];
	  I[slice(d, Nex, ddim)] = b && a;
          valarray<int> temp4 = amarg[slice(d, dim2, ddim)];
          temp4[k] = ave[k] + 1;
          amarg[slice(d, dim2, ddim)] = temp4;
        } else {
  	  valarray<bool> a = J[slice(k, Nex, ave.size())];
	  valarray<bool> b = I[slice(d, Nex, ddim)];
          I[slice(d, Nex, ddim)] = b && a;
          valarray<int> temp4 = amarg[slice(d, dim2, ddim)];
          temp4[k + dim] = ave[k];
          amarg[slice(d, dim2, ddim)] = temp4;
        }
      }
    }

    valarray<int> truth(1, Nex);
    for (int i = 0; i < ddim; i++) {
      valarray<bool> temp4 = I[slice(i, Nex, ddim)];
      valarray<int> th = truth[temp4==true];
      NN[i] = (double)th.sum();
    }

    double constant = ((double) Nex) / ddim;
    valarray<double> difference = NN - constant;
    valarray<double> squared = pow(difference, 2.0);
    double sum = squared.sum();
    double tst = (sum * ddim) / ((double) Nex);

    if (tst > chi2[dim - 1] || run == 1) {
       npar--;
       for (int i = 0; i < ddim; i++) {
         if (NN[i] > ddim) {
           ++npar;
           akon = apoc + (int) NN[i] - 1;
           poc[npar - 1] = apoc;
           kon[npar - 1] = akon;

           valarray<int> temp = amarg[slice(i, dim2, ddim)];
    	   marg[slice(npar - 1, dim2, (8 * ddim))] = temp; 

           valarray<bool> indices = I[slice(i, Nex, ddim)];
           valarray<int> t = valarray<int>(apor[indices==true]);
           poradi[slice(apoc - 1, akon - apoc + 1, 1)] = t;
           apoc = akon + 1;
         } else if (NN[i] > 0) {
           valarray<int> row_i = amarg[slice(i, dim2, ddim)];
           double Nz = (double)(row_i[5] - row_i[2] + 1);
	   valarray<int> a = row_i[slice(dim, dim2 - dim, 1)];
	   valarray<int> b = row_i[slice(0, dim, 1)];
	   valarray<int> c = valarray<int>(1, dim);
	   valarray<int> temp5 = a - b + c;
           double Nxx = 1;
           for (uint te = 0; te < temp5.size(); te++) {
             Nxx *= (double) temp5[te];
           }

	   valarray<int> a0 = ranks[slice(0, N, dim)];
	   valarray<bool> Jx = (a0 >= row_i[0]) && (a0 <= row_i[3]);
	   valarray<int> a1 = ranks[slice(1, N, dim)];
	   valarray<bool> Jy = (a1 >= row_i[1]) && (a1 <= row_i[4]);
	   valarray<int> a2 = ranks[slice(2, N, dim)];
	   valarray<bool> Jz = (a2 >= row_i[2]) && (a2 <= row_i[5]);
	   valarray<bool> temp6 = Jx && Jz;
	   valarray<bool> temp7 = Jy && Jz;

	   double Nxz = 0;
	   for (uint c = 0; c < temp6.size(); c++) {
             if (temp6[c])
               Nxz++;
           }
           double Nyz = 0;
           for (uint c = 0; c < temp7.size(); c++) {
             if (temp7[c])
                Nyz++;
           }
           double cond = 1;
           if (((NN[i] * Nz) != 0) && ((Nxz * Nyz) != 0)) {
              cond = (NN[i] * Nz) / (Nxz * Nyz);
	   }
           xcor += NN[i] * log(cond);
         }
       }
   } else {
       valarray<int> row_npar = marg[slice(npar - 1, dim2, (8 * ddim))];
       double Nz = row_npar[5] - row_npar[2] + 1;
       valarray<int> a = row_npar[slice(dim, dim2 - dim, 1)];
       valarray<int> b = row_npar[slice(0, dim, 1)];
       valarray<int> c = valarray<int>(1, dim); 
       double Nxx = 1;
       valarray<int> abc = a - b + c;
       for (uint te = 0; te < abc.size(); te++) {
	 Nxx *= abc[te];
       }

       valarray<int> a0 = ranks[slice(0, N, dim)];
       valarray<bool> Jx = (a0 >= row_npar[0]) && (a0 <= row_npar[3]);
       valarray<int> a1 = ranks[slice(1, N, dim)]; 
       valarray<bool> Jy = (a1 >= row_npar[1]) && (a1 <= row_npar[4]);
       valarray<int> a2 = ranks[slice(2, N, dim)];
       valarray<bool> Jz = (a2 >= row_npar[2]) && (a2 <= row_npar[5]);
        
       valarray<bool> temp6 = Jx && Jz;
       valarray<bool> temp7 = Jy && Jz;

       double Nxz = 0;
       for (uint c = 0; c < temp6.size(); c++) {
         if (temp6[c])
            Nxz++;
       }

       double Nyz = 0;
       for (uint c = 0; c < temp7.size(); c++) {
         if (temp7[c])
            Nyz++;
       }

       double cond = 1;
       if (((Nex * Nz) != 0) && ((Nxz * Nyz) != 0)) {
          cond = (Nex * Nz) / (Nxz * Nyz);
       }
       xcor += Nex * log(cond);
       npar--;
     }
   }
   xcor /= N;
   return xcor;
}

void Set_Triplet( Gene_Triplet & p, double xId, double yId, double zId, int maId )
{
   p.Set_X( xId );
   p.Set_Y( yId );
   p.Set_Z( zId );
   p.Set_MaID( maId );
}
