//
// Copyright (C) 2003  Columbia Genome Center
// All Rights Reserved.
//
// MutualInfo.h -- Class and constant definitions for mutual
// information calculation progoram.
//
// $Id: MutualInfo.h,v 1.2 2013/03/16 22:01:18 manju Exp $
//

/******************************************************************************
* Class MutualInfo
*
* This class contains methods for computing mutual information between a pair
* of genes
*
* Current implementation supports two type of estimations
* 1) fast - binning
* 2) accurrate - gaussian kernel estimator
*
* Inner Class Gene_Pair
*             Sort_X
*             Sort_Y
*
* are the data structure a MutualInfo object takes to compute the MI, and
* functions used to prepare the two vectors of gene expressions.
*
*******************************************************************************
*/


#ifndef MUTUALINFO_H__
  #define MUTUALINFO_H__
  #include <vector>
  #include "loki/TypeManip.h"
  #include <string.h>

using namespace std;
typedef unsigned int uint;

/****************************************************************************
* Auxiliary classes for sorting sets of gene pairs.
*
**/
class Gene_Pair
{

    double x;
    double y;
    int xi; // index of x
    int yi; // index y
    int maId;

public:

    inline double Get_X() const
    {
        return x;
    }

    inline double Get_Y() const
    {
        return y;
    }

    inline int Get_XI() const
    {
        return xi;
    }

    inline int Get_YI() const
    {
        return yi;
    }

    inline int Get_MaID() const
    {
        return maId;
    }

    inline void Set_X( double X )
    {
        x = X;
    }

    inline void Set_Y( double Y )
    {
        y = Y;
    }

    inline void Set_XI( int XI )
    {
        xi = XI;
    }

    inline void Set_YI( int YI )
    {
        yi = YI;
    }

    inline void Set_MaID( int MaID )
    {
        maId = MaID;
    }
};


typedef std::vector < Gene_Pair > Pair_Vector;


class Sort_X : std::binary_function < Gene_Pair, Gene_Pair, bool >
{

public:

    bool operator() ( const Gene_Pair & a, const Gene_Pair & b ) const
    {
        if ( a.Get_X() != b.Get_X() )
        {
            return ( a.Get_X() < b.Get_X() );
        }
        else
        {
            return ( a.Get_MaID() < b.Get_MaID() );
        }
    }
};


class Sort_Y : std::binary_function < Gene_Pair, Gene_Pair, bool >
{

public:

    bool operator() ( const Gene_Pair & a, const Gene_Pair & b ) const
    {
        if ( a.Get_Y() != b.Get_Y() )
        {
            return ( a.Get_Y() < b.Get_Y() );
        }
        else
        {
            return ( a.Get_MaID() < b.Get_MaID() );
        }
    }
};

/****************************************************************************
* Auxiliary classes for sorting sets of gene triplets.
*
**/
class Gene_Triplet
{

    double x;
    double y;
    double z;
    int xi; // index of x
    int yi; // index y
    int zi; // index of z
    int maId;

public:

    inline double Get_X() const
    {
        return x;
    }

    inline double Get_Y() const
    {
        return y;
    }

    inline double Get_Z() const
    {   
        return z;
    }   

    inline int Get_XI() const
    {
        return xi;
    }

    inline int Get_YI() const
    {
        return yi;
    }

    inline int Get_ZI() const
    {
        return zi;
    }

    inline int Get_MaID() const
    {
        return maId;
    }

    inline void Set_X( double X )
    {
        x = X;
    }

    inline void Set_Y( double Y )
    {
        y = Y;
    }

    inline void Set_Z( double Z )
    {
        z = Z;
    }

    inline void Set_XI( int XI )
    {
        xi = XI;
    }

    inline void Set_YI( int YI )
    {
        yi = YI;
    }

    inline void Set_ZI( int ZI )
    {
        zi = ZI;
    }

    inline void Set_MaID( int MaID )
    {
        maId = MaID;
    }
};


typedef std::vector < Gene_Triplet > Triplet_Vector;


class Sort_XT : std::binary_function < Gene_Triplet, Gene_Triplet, bool >
{

public:

    bool operator() ( const Gene_Triplet & a, const Gene_Triplet & b ) const
    {
        if ( a.Get_X() != b.Get_X() )
        {
            return ( a.Get_X() < b.Get_X() );
        }
        else
        {
            return ( a.Get_MaID() < b.Get_MaID() );
        }
    }
};


class Sort_YT : std::binary_function < Gene_Triplet, Gene_Triplet, bool >
{

public:

    bool operator() ( const Gene_Triplet & a, const Gene_Triplet & b ) const
    {
        if ( a.Get_Y() != b.Get_Y() )
        {
            return ( a.Get_Y() < b.Get_Y() );
        }
        else
        {
            return ( a.Get_MaID() < b.Get_MaID() );
        }
    }
};

class Sort_ZT : std::binary_function < Gene_Triplet, Gene_Triplet, bool >
{

public:

    bool operator() ( const Gene_Triplet & a, const Gene_Triplet & b ) const
    {
        if ( a.Get_Z() != b.Get_Z() )
        {
            return ( a.Get_Z() < b.Get_Z() );
        }
        else
        {
            return ( a.Get_MaID() < b.Get_MaID() );
        }
    }
};


/****************************************************************************
* Class MutualInfo
*
* The main MutualInfo class
*
**/
class Mutual_Info
{
    static const int MIBLOCKS = 2;
    static const int BINS = 1000;

    uint     Microarray_Num;
    int     miSteps;
    int     ** MI_Space;
    double  ** MI_Prob;
    double  MA_Per_MI_Step;
    double  * histogram;
    double  * background;
    int     Max_Histo_Bin;
    int     Max_Background_Bin;

    //type of caluclation
    int type;

    // variance2 = 2*sigma*sigma, doing this just same some computation
    double variance2;
    double ** prob_table;
    double *  norm_1D_table;
    double ** norm_2D_table;
    bool Copula_Transform;

public:

    static int const RANK_NONE = 0;
    static int const PEARSON = 1;
    static int const SPEARMAN = 2;
    static int const MI_GAUSSIAN = 3;
    static int const MI_AVERAGE = 4;
    static int const MI_ADAPTIVE_PARTITIONING = 5;
    static int const CMI_AP = 6;

    Mutual_Info( int, int, double, int );
    ~Mutual_Info();
    int Get_Type() const
    {
        return type;
    }

    void Set_Copula_Transform( bool t = true )
    {
        Copula_Transform = t;
    }
    bool Is_Copula_Transform() const
    {
        return Copula_Transform;
    }
    double Compute_Pairwise_MI( Pair_Vector & );
    uint Get_Num_Microarrays()
    {
        return Microarray_Num;
    }

    double Get_Conditional_Mutual_Info_XYZ( Triplet_Vector & );
private:

    void Initialize_Norm_Table( double );
    double Get_Kernel( Pair_Vector & pairs, unsigned int i );
    inline void randomize_X(Gene_Pair & gp);
    inline void randomize_Y(Gene_Pair & gp);
    double Get_Mutual_Info_XY( Pair_Vector &, const Loki::Int2Type < Mutual_Info::MI_AVERAGE > & );
    double Get_Mutual_Info_XY( Pair_Vector &, const Loki::Int2Type < Mutual_Info::MI_GAUSSIAN > & );
    double Get_Mutual_Info_XY( Pair_Vector &, const Loki::Int2Type < Mutual_Info::PEARSON > & );
    double Get_Mutual_Info_XY( Pair_Vector &, const Loki::Int2Type < Mutual_Info::SPEARMAN > & );
    double Get_Mutual_Info_XY( Pair_Vector & pairs, const Loki::Int2Type < Mutual_Info::MI_ADAPTIVE_PARTITIONING > & );

};

#endif
