

/******************************************************************************
 *
 *  file:  BoundaryConstraint.h
 *
 *  Copyright (c) 2005, Michael E. Smoot
 *  All rights reverved.
 *
 *  See the file COPYING in the top directory of this distribution for
 *  more information.
 *
 *  THE SOFTWARE IS PROVIDED _AS IS_, WITHOUT WARRANTY OF ANY KIND, EXPRESS
 *  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 *  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 *
 *****************************************************************************/

#ifndef TCLAP_BOUNDARYCONSTRAINT_H
#define TCLAP_BOUNDARYCONSTRAINT_H

#include <string>
#include <vector>
#include <tclap/Constraint.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
#define HAVE_SSTREAM
#endif

#if defined(HAVE_SSTREAM)
#include <sstream>
#elif defined(HAVE_STRSTREAM)
#include <strstream>
#else
#error "Need a stringstream (sstream or strstream) to compile!"
#endif

namespace TCLAP
{

    /**
     * A Constraint that constrains the Arg to only those values specified
     * in the constraint.
     */
    template<class T>
    class BoundaryConstraint : public Constraint<T>
    {

    public:

        /**
         * Constructor.
         * \param lower bound
         */
        BoundaryConstraint(T boundary, bool lower, bool incl, std::string type);

        /**
         * Virtual destructor.
         */
        virtual ~BoundaryConstraint()
        {}

        /**
         * Returns a description of the Constraint.
         */
        virtual std::string description() const;

        /**
         * Returns the short ID for the Constraint.
         */
        virtual std::string shortID() const;

        /**
         * The method used to verify that the value parsed from the command
         * line meets the constraint.
         * \param value - The value that will be checked.
         */
        virtual bool check(const T& value) const;

    protected:

        T _boundary;
        bool _lower;
        bool _incl;
        std::string _type;

        /**
         * The string used to describe the allowed values of this constraint.
         */
        std::string _typeDesc;

    };

    template<class T>
    BoundaryConstraint<T>::BoundaryConstraint(T boundary, bool lower, bool incl, std::string type)
            : _boundary(boundary),_lower(lower),_incl(incl),_type(type)
    {

#if defined(HAVE_SSTREAM)
        std::ostringstream os;
#elif defined(HAVE_STRSTREAM)

        std::ostrstream os;
#else
#error "Need a stringstream (sstream or strstream) to compile!"
#endif

        if (_lower)
        {
            if (_incl)
            {
                os << "[" << _boundary << ", )";
            }
            else
            {
                os << "(" << _boundary << ", )";
            }
        }
        else
        {
            if (_incl)
            {
                os << "( ," << _boundary << "]";
            }
            else
            {
                os << "( ," << _boundary << ")";
            }
        }
        _typeDesc = os.str();
    }


    template<class T>
    bool BoundaryConstraint<T>::check( const T& val ) const
    {
        if (_lower)
        {
            if (_incl)
            {
                return (val >= _boundary);
            }
            else
            {
                return (val > _boundary);
            }
        }
        else
        {
            if (_incl)
            {
                return (val <= _boundary);
            }
            else
            {
                return (val < _boundary);
            }
        }
    }


    template<class T>
    std::string BoundaryConstraint<T>::shortID() const
    {
        return _type;
    }

    template<class T>
    std::string BoundaryConstraint<T>::description() const
    {
        return _typeDesc;
    }


} //namespace TCLAP
#endif

