

/******************************************************************************
 *
 *  file:  RangeConstraint.h
 *
 *  Copyright (c) 2005, Michael E. Smoot
 *  All rights reverved.
 *
 *  See the file COPYING in the top directory of this distribution for
 *  more information.
 *
 *  THE SOFTWARE IS PROVIDED _AS IS_, WITHOUT WARRANTY OF ANY KIND, EXPRESS
 *  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 *  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 *  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 *  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
 *  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
 *  DEALINGS IN THE SOFTWARE.
 *
 *****************************************************************************/

#ifndef TCLAP_RANGECONSTRAINT_H
#define TCLAP_RANGECONSTRAINT_H

#include <string>
#include <vector>
#include <tclap/Constraint.h>

#ifdef HAVE_CONFIG_H
#include <config.h>
#else
#define HAVE_SSTREAM
#endif

#if defined(HAVE_SSTREAM)
#include <sstream>
#elif defined(HAVE_STRSTREAM)
#include <strstream>
#else
#error "Need a stringstream (sstream or strstream) to compile!"
#endif

namespace TCLAP
{

    /**
     * A Constraint that constrains the Arg to only those values specified
     * in the constraint.
     */
    template<class T>
    class RangeConstraint : public Constraint<T>
    {

    public:

        /**
         * Constructor.
         * \param lower bound
         */
        RangeConstraint(T lower_bound, bool lower_incl, T upper_bound, bool upper_incl, std::string type);

        /**
         * Virtual destructor.
         */
        virtual ~RangeConstraint()
        {}

        /**
         * Returns a description of the Constraint.
         */
        virtual std::string description() const;

        /**
         * Returns the short ID for the Constraint.
         */
        virtual std::string shortID() const;

        /**
         * The method used to verify that the value parsed from the command
         * line meets the constraint.
         * \param value - The value that will be checked.
         */
        virtual bool check(const T& value) const;

    protected:

        T _lower_bound;
        bool _lower_incl;
        T _upper_bound;
        bool _upper_incl;
        std::string _type;

        /**
         * The string used to describe the allowed values of this constraint.
         */
        std::string _typeDesc;

    };

    template<class T>
    RangeConstraint<T>::RangeConstraint(T lower_bound, bool lower_incl,
                                        T upper_bound, bool upper_incl,std::string type)
            : _lower_bound(lower_bound),_lower_incl(lower_incl),
            _upper_bound(upper_bound),_upper_incl(upper_incl),_type(type)
    {

#if defined(HAVE_SSTREAM)
        std::ostringstream os;
#elif defined(HAVE_STRSTREAM)

        std::ostrstream os;
#else
#error "Need a stringstream (sstream or strstream) to compile!"
#endif

        if (_lower_incl)
        {
            os << "[";
        }
        else
        {
            os << "(";
        }

        os << _lower_bound << "," << _upper_bound;

        if (_upper_incl)
        {
            os << "]";
        }
        else
        {
            os << ")";
        }

        _typeDesc = os.str();
    }


    template<class T>
    bool RangeConstraint<T>::check( const T& val ) const
    {
        bool lower_satisfied;
        bool upper_satisfied;

        if (_lower_incl)
        {
            lower_satisfied = (val >= _lower_bound);
        }
        else
        {
            lower_satisfied = (val > _lower_bound);
        }

        if (_upper_incl)
        {
            upper_satisfied = (val <= _upper_bound);
        }
        else
        {
            upper_satisfied = (val < _upper_bound);
        }

        if ( lower_satisfied && upper_satisfied )
            return true;
        else
            return false;
    }


    template<class T>
    std::string RangeConstraint<T>::shortID() const
    {
        return _type;
    }

    template<class T>
    std::string RangeConstraint<T>::description() const
    {
        return _typeDesc;
    }


} //namespace TCLAP
#endif

