//
// Copyright (C) 2003 Columbia Genome Center
// All Rights Reserved.
//
// main.cc -- Main routines.
// vi: set ts=4
//
// $Id: main.cpp,v 1.4 2013/03/16 22:01:08 manju Exp $
//

#include <map>
#include <string>
#include <list>
#include "Microarray_Set.h"
#include "MutualInfo.h"
#include "CMatrix.h"
#include "param.h"
#include "tclap/CmdLine.h"

using namespace TCLAP;
using namespace std;
typedef unsigned int uint;

Parameter parseOption( int argc, char * argv[] )
{
    Parameter p;
    try
    {
        string ver = "2.0, 10/28/2007";
        string mesg = "OUTPUTS: .cmm (Conditional Mutual Information Matrix) files\n";
        CmdLine cmd(mesg,' ',ver);

        ValueArg<string> infile_arg("i","gep_dataset","Input Gene Expression dataset",true,"","file path");
        cmd.add( infile_arg );

        ValueArg<string> candidate_arg("c","gep_modulator","Modulator Gene Expression dataset",true,"","file path");
        cmd.add( candidate_arg );

        ValueArg<string> hub_arg("u","hub_id","Hub gene probe Id",true,"","string");
        cmd.add( hub_arg );

        ValueArg<string> outfile_arg("o","output_file","Output file path/name",false,"","file path");
        cmd.add( outfile_arg );

        ValueArg<string> tf_arg("l","TF_list","List of transcription factor (TF) probe Ids",false,"","file path");
        cmd.add( tf_arg );

        vector <string> allowed;
        allowed.push_back("pearson");
        allowed.push_back("spearman");
        allowed.push_back("fixed_bandwidth");
        allowed.push_back("fast");
	allowed.push_back("adaptive_partitioning");
        allowed.push_back("cmi_ap");
        ValuesConstraint<string> allowed_algo( allowed );
        ValueArg<string> algo_arg("a","algorithm","MI estimator",false,"fixed_bandwidth",&allowed_algo);
        cmd.add( algo_arg );

        ValueArg<string> home_arg("H","directory","Directory containing configuration files",false,"./","dir");
        cmd.add( home_arg );

        RangeConstraint<float> range(0,true,1,true,"float");
        ValueArg<float> kernel_s_arg("k","kernel_subset","Kenel width (subset)",false,0,&range);
        cmd.add( kernel_s_arg );

        ValueArg<float> kernel_f_arg("K","kernel_full","Kernel width (full dataset)",false,0,&range);
        cmd.add( kernel_f_arg );

        BoundaryConstraint<float> bound(0,true,true,"float");
        ValueArg<float> threshold_s_arg("t","threshold_subset","MI threshold (subset)",false,0,&bound);
        cmd.add( threshold_s_arg );

        ValueArg<float> threshold_f_arg("T","threshold_full","MI threshold (full dataset)",false,10,&bound);
        cmd.add( threshold_f_arg );

        ValueArg<float> pvalue_s_arg("p","pvalue_subset","MI p-value (subset)",false,1,&range);
        cmd.add( pvalue_s_arg );

        ValueArg<float> pvalue_f_arg("P","pvalue_full","MI p-value (full dataset)",false,1,&range);
        cmd.add( pvalue_f_arg );

        ValueArg<float> tolerance_arg("e","tolerance","DPI tolerance",false,1,&range);
        cmd.add( tolerance_arg );

        ValueArg<float> percent_arg("q","percent","Percentage tail",false,0.25,&range);
        cmd.add( percent_arg );

        ValueArg<float> mean_arg("m","mean","Minimal mean expression value",false,0,&bound);
        cmd.add( mean_arg );

        ValueArg<float> cv_arg("s","cv","Minimal coefficient of variation value",false,0,&bound);
        cmd.add( cv_arg );

        SwitchArg verbose_switch("v","verbose","Display parameter values", false);
        cmd.add( verbose_switch );

        cmd.parse( argc, argv );

        p.infile = infile_arg.getValue();
        p.outfile = outfile_arg.getValue();
        p.hub = hub_arg.getValue();
        p.candidate_file = candidate_arg.getValue();
        p.tf_list_file = tf_arg.getValue();
        p.mean = mean_arg.getValue();
        p.cv = cv_arg.getValue();
        p.threshold = threshold_s_arg.getValue();
        p.Threshold = threshold_f_arg.getValue();
        p.eps = tolerance_arg.getValue();
        p.sigma = kernel_s_arg.getValue();
        p.Sigma = kernel_f_arg.getValue();
        p.pvalue = pvalue_s_arg.getValue();
        p.Pvalue = pvalue_f_arg.getValue();
        p.percent = percent_arg.getValue();
        p.verbose = verbose_switch.getValue();
        p.algorithm = algo_arg.getValue();
        p.home_dir = home_arg.getValue();
    }
    catch (ArgException &e)  // catch any exceptions
    {
        cerr << "error: " << e.error() << " for arg " << e.argId() << endl;
    }
    return p;
}


int main( int argc, char * argv[] )
{
    try
    {
        Parameter p = parseOption( argc, argv );

        Microarray_Set data;
        Microarray_Set candidate;

        processParameter(p, data, candidate);

        int TYPE;
        if (cmpIgnoreCase( p.algorithm.c_str(), "adaptive_partitioning" )==0)
        {
            TYPE = Mutual_Info::MI_ADAPTIVE_PARTITIONING;
        }
        else if (cmpIgnoreCase( p.algorithm.c_str(), "fixed_bandwidth" )==0)
        {
            TYPE = Mutual_Info::MI_GAUSSIAN;
        }
        else if (cmpIgnoreCase( p.algorithm.c_str(), "fast" )==0)
        {
            TYPE = Mutual_Info::MI_AVERAGE;
        }
        else if (cmpIgnoreCase( p.algorithm.c_str(), "pearson" )==0)
        {
            TYPE = Mutual_Info::PEARSON;
        }
        else if (cmpIgnoreCase( p.algorithm.c_str(), "cmi_ap" )==0)
        {
	    TYPE = Mutual_Info::CMI_AP;
	}
        else
        {
            TYPE = Mutual_Info::SPEARMAN;
        }

        map <int, int> transfact;
        if (not(p.tflist.empty()))
        {
            for (uint i=0; i<p.tflist.size(); i++)
            {
                transfact[p.tflist[i]] = i;
            }
        }

        CMatrix cmatrix;
        // unconditional Mutual Information calculator
        Mutual_Info mi(p.nsample, p.miSteps, p.Sigma, TYPE );
        // Conditional Mutual Information calculator
        Mutual_Info cmi(p.nsubsample, p.miSteps, p.sigma, TYPE );
        cmatrix.computeCondMatrix(data,candidate,mi,cmi,p.hubid,p.threshold,p.eps,p.percent,transfact);
        cmatrix.write(data, candidate, p.hubid, p.outfile+".cmm",true, TYPE);
        //if ( (p.threshold != 0) || (p.eps != 1) )
        //{
            //cmatrix.write(data, candidate, p.hubid, p.outfile+".cam",true);
        //}
    }
    catch ( string & s )
    {
        cout << s << endl;
    }

    return ( 0 );
}
