/*
 * Copyright (C) 2003, 2004  Columbia Genome Center
 * All Rights Reserved.
 *
 * param.cc --
 *
 * $Id: param.cpp,v 1.3 2010/03/10 22:05:13 manju Exp $
 */

#include <fstream>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <string.h>
#include <cmath>
#include "param.h"

using namespace std;

int cmpIgnoreCase(const char* a, const char* b)
{
    int i = 0;
    for(; toupper(a[i])==toupper(b[i]); ++i)
    {
        if(a[i] == '\0')
            return 0;
    }
    return(a[i]-b[i]);
}


double findKernelWidth(int n, string base)
{
    string filename = base+"config_kernel.txt";
    ifstream infile(filename.c_str(), ifstream::in);

    if(!infile.good())
    {
        throw(string("Cannot find file: "+filename+"! Please use '-H' option to specify the ARACNE home directory."));
    }

    string line;
    double alpha;
    double beta;
    getline(infile, line);
    while (strncmp(line.c_str(),">",1) == 0)
    {
        getline(infile, line);
    }

    istringstream sin(line);
    if ( sin.good() && (sin.peek() != EOF) )
    {
        sin >> alpha >> beta;
    }
    else
    {
        throw(string("Configuration file format error: "+filename));
    }
    infile.close();
    return alpha*pow(n,beta);
}


double findThreshold(int n, double pvalue, string base)
{
    string filename = base+"config_threshold.txt";
    ifstream infile(filename.c_str(), ifstream::in);

    if(!infile.good())
    {
        throw(string("Cannot find file: "+filename+"! Please use '-H' option to specify the ARACNE home directory."));
    }

    string line;
    double alpha;
    double beta;
    double gamma;
    getline(infile, line);
    while (strncmp(line.c_str(),">",1) == 0)
    {
        getline(infile, line);
    }

    istringstream sin(line);
    if ( sin.good() && (sin.peek() != EOF) )
    {
        sin >> alpha >> beta >> gamma;
    }
    else
    {
        throw(string("Configuration file format error: "+filename));
    }
    infile.close();
    return (alpha - log(pvalue))/((-beta)+(-gamma)*n);
}


string getFileExt(string matrixName)
{
    unsigned int  len = matrixName.length();
    string ext = "";
    string::size_type c = matrixName.find_last_of(".");
    if(c != matrixName.npos)
    {
        ext = matrixName.substr(c+1,len-1);
    }
    return ext;
}


std::string getFileName(string filename)
{
    unsigned int  len = filename.length();
    string::size_type b   = filename.find_last_of("/");
    if(b == std::string::npos)
    {
        b = filename.find_last_of("\\");
    }
    string dirname("");
    string basename(filename);
    if ((b != filename.npos) && (b < len))
    {
        // Extract the directory and the filename if path is included
        basename = filename.substr(b + 1, len);
        dirname = filename.substr(0, b) + string("/");
    }
    string::size_type c = basename.find_last_of(".");
    if(c != basename.npos)
    {
        basename = basename.substr(0,c);
    }
    string name = dirname + basename;
    return name;
}


void processParameter(Parameter &p, Microarray_Set & data, Microarray_Set & candidate)
{
    // **************************************************************************
    //            Load data
    // **************************************************************************
    data.read(p.infile);
    p.nmarker = data.Get_Num_Markers();
    p.nsample = data.Get_Num_Microarrays();
    // round to the nearest integer
    p.nsubsample = (int)((double)(p.nsample)* p.percent + 0.5);

    candidate.read(p.candidate_file);
    p.ncandidate = candidate.Get_Num_Markers();

    p.hubid = data.Get_Marker_Id( p.hub );
    if (p.hubid == -1)
    {
        throw(string("Cannot find hub: "+p.hub+"!"));
    }

    if (p.tf_list_file != "")
    {
        readProbeList(p.tf_list_file, p.tflist, data);
    }
    // **************************************************************************
    //           Prepare MI parameter for all samples
    // **************************************************************************
    if (p.Sigma == 0)
    {
        p.Sigma = findKernelWidth( p.nsample, p.home_dir );
    }

    if ( p.Threshold==10 && p.Pvalue!=1 )
    {
        p.Threshold = findThreshold( p.nsample, p.Pvalue, p.home_dir );
    }

    // **************************************************************************
    //            Prepare MI parameter for conditional MI
    // **************************************************************************
    if (p.sigma == 0)
    {
        p.sigma = findKernelWidth( p.nsubsample, p.home_dir );
    }

    if ( p.threshold==0 && p.pvalue!=1 )
    {
        p.threshold = findThreshold( p.nsubsample, p.pvalue, p.home_dir );
    }

    // **************************************************************************
    //            Prepare output file
    // **************************************************************************

    if (p.outfile == "")
    {
        string filename = getFileName(p.infile);
        string temp;

        if (p.hub !="")
        {
            filename = filename+"_h"+p.hub;
        }

        if (cmpIgnoreCase(p.algorithm.c_str(),"fixed_bandwidth" )==0)
        {
            char kernel[10];
            sprintf(kernel, "%0.3g", p.sigma);
            filename = filename+"_k"+kernel;

            if (p.threshold>0)
            {
                char threshold[10];
                sprintf(threshold, "%0.3g", p.threshold);
                filename = filename+"_t"+threshold;
            }

            if (p.eps < 1)
            {
                char eps[10];
                sprintf(eps, "%0.2g", p.eps);
                filename = filename+"_e"+eps;
            }
        }

        char pcnt[10];
        sprintf(pcnt, "%0.2g", p.percent);
        filename = filename+"_p"+pcnt;

        p.outfile = filename;
    }
    // *************************************************************************
    //            Prepare for using PEARSON/SPEARMAN correlation
    // *************************************************************************
    if ( (cmpIgnoreCase(p.algorithm.c_str(),"pearson" )==0) || (cmpIgnoreCase(p.algorithm.c_str(),"spearman" )==0) )
    {
        p.eps = 1;
        cout << "DPI is disabled when PEARSON/SPEARMAN correlation is used." << endl;
    }
    if (p.verbose) {
        displayParameter(p);
    }
}


void displayParameter(Parameter & p)
{
    // **************************************************************************
    //            Display parameters
    // **************************************************************************
    cout << endl;
    cout << "[PARA] Input file          :" << p.infile << endl;
    cout << "[PARA]     No. marker      :" << p.nmarker << endl;
    cout << "[PARA]     No. sample      :" << p.nsample << endl;
    cout << endl;
    cout << "[PARA] Output file         :" << p.outfile << endl;
    cout << endl;
    cout << "[PARA] Algorithm           :" << p.algorithm << endl;
    cout << endl;
    cout << "[PARA] Hub gene:           :" << p.hub << " (" << p.hubid << ")" << endl;
    cout << endl;
    cout << "[PARA] Candidate gene list :" << p.candidate_file << " (" << p.ncandidate << ")" << endl;
    cout << "[PARA] Condition size      :" << p.percent << " (" << p.nsubsample << ")" << endl;
    cout << endl;
    if (cmpIgnoreCase(p.algorithm.c_str(),"fixed_bandwidth" )==0)
    {
        cout << "[PARA] MI kernel width " << endl;
        cout << "[PARA]     All sample      :" << p.Sigma << endl;
        cout << "[PARA]     Conditonal      :" << p.sigma << endl;
        cout << endl;
        cout << "[PARA] MI Threshold " << endl;
        cout << "[PARA]     All sample      :" << p.Threshold << endl;
        cout << "[PARA]     Conditonal      :" << p.threshold << endl;
        cout << endl;
    }
    cout << "[PARA] DPI tolerance       :" << p.eps << endl;
    cout << endl;
    cout << "[PARA] TF annotation list  :" << p.tf_list_file <<" (" << p.tflist.size() << ")" << endl;
    cout << endl;
}

void readProbeList(string file, vector<int> & list, Microarray_Set & data)
{
    ifstream in(file.c_str());
    if (!in.good())
    {
        throw(std::string("Problem with reading the file "+file));
    }
    string line;
    while ( in.good() && (in.peek() != EOF) && (in.peek() != '\012') )
    {
        getline(in, line);
        istringstream sin(line);
        string token;
        sin >> token;

        int gid = data.Get_Marker_Id( token );
        if (gid == -1)
        {
            throw(string("Missing probe: "+token+" in "+file));
        }
        list.push_back(gid);
    }
}
