function postProcessMINDYCluster(hub,gep_file,indir,outdir,tempdir)

if strcmp(indir(end),'/')==0
    indir = [indir '/'];
end

if strcmp(outdir(end),'/')==0
    outdir = [outdir '/'];
end

if strcmp(tempdir(end),'/')==0
    tempdir = [tempdir '/'];
end

% algorithm parameters
percent     = 0.35; % percentage tail
p_independ  = 1e-7; % MI p-value for independence test of candidate modulators
p_dynamics  = 1;    % statistical significance for tail separation
p_discover  = 0.05; % family-wise statistical significance for conditional mutual information difference
p_mode      = 0.10;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Load data and MINDY results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
disp('Loading data ... ');
load(gep_file);

disp('Reading MINDY results ... ');
cmm_file = ['data_h' hub '_k0.179_t0.164_e0.1_p0.35.cmm'];
cam_file = ['data_h' hub '_k0.179_t0.164_e0.1_p0.35.cam'];

if exist([indir cam_file],'file')==0
    if exist([indir cam_file '.bz2'],'file')~=0
        disp('--> decompressing ...');
        temp_cam_file = [tempdir cam_file];
        unix(['bzip2 -d -c ' indir cam_file '.bz2 > ' temp_cam_file]);        
    else
        error(['Missing CAM file: ' hub]);
    end
else
    temp_cam_file = [indir cam_file];
end

if exist([indir cmm_file],'file')==0
    if exist([indir cmm_file '.bz2'],'file')~=0
        disp('--> decompressing ...');
        temp_cmm_file = [tempdir cmm_file];
        unix(['bzip2 -d -c ' indir cmm_file '.bz2 > ' temp_cmm_file]);        
    else
        error(['Missing CMM file: ' hub]);
    end
else
    temp_cmm_file = [indir cmm_file];
end

[edges,foo]=textread(temp_cam_file,'%s%[^\n]','delimiter','\t','headerlines',1,'bufsize', 65520);
% NOTE: the second column is the unconditional mutual information
[foo,bar,header] = textread(temp_cam_file,'%s%s%[^\n]',1,'delimiter','\t','bufsize', 65520);
conditions = strread(header{1},'%s','delimiter','\t');

if (mod(length(conditions),2) ~= 0)
    error('Conditional matrix format error!');
end

candidates = cell(length(conditions)/2,1);
for j=1:2:length(conditions)
    [gene cond] = strtok(conditions{j},':');
    candidates{(j+1)/2} = gene;
end

% the first column of both matrices is the unconditional mutual information
temp    = single(dlmread(temp_cam_file,'\t',1,1));
mi      = temp(:,1);
cam     = temp(:,2:size(temp,2));
clear temp;
cmm     = single(dlmread(temp_cmm_file,'\t',1,2));

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Select candidate modulators
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
% make sure that rows in CAM are in the same order as those in the EXP
[Z,IX,IY] = intersect(gep.affyid,edges);
if sum(IX-(1:length(IX))')~=0   % IX should be a vector of consecutive integers
    error('Data format does not match!');
end
% get candidates indices in the expression dataset
[C,IA,IB] = intersect(candidates,edges);
if length(C)~=length(candidates)    % {candidates} is a subset of {edges}
    error('Missing probes!');
end
[IC,I] = sort(IA);
% 1) by dynamics - expression values of the candidates
by_dynamics = hasSeparatedTails(gep.data(IB(I),:),percent,p_dynamics);
% 2) by indepedence - unconditional mutual information of the candidates
% with the hub
% p-value ~ 0.05/[no. candidates]
by_independ = (mi(IB(I)) <= getMIThreshold(254,p_independ));  
% indices of candidate modulator
cidx = find(by_dynamics & by_independ);

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Select candidate targets as those supported under at least one
%   condition
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
ridx = find(sum(cam,2)); % row sum

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Compute conditional mutual information score
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
candidate_modulator = candidates(cidx);
candidate_target    = edges(ridx);
% mutual information of M+ minus that of M-
score       = cmm(ridx,cidx.*2) - cmm(ridx,cidx.*2-1);
max_score   = max(cmm(ridx,cidx.*2), cmm(ridx,cidx.*2-1));
direct      = (cam(ridx,cidx.*2) | cam(ridx,cidx.*2-1));
clear cam cmm;
% if there is no edge under either conditions of a candidate modulator, set
% the conditional mutual information difference to be zero (i.e.
% insignificant).
score   = score.*direct;
% p-value of a conditional mutual information difference depends on the
% unconditional mutual information of the interaction.
pv      = pvalue(mi(ridx),abs(score));
% record triplet removed due to DPI
idx     = find(direct==0);
pv(idx) = NaN;

% Bonferroni correction for all triplet examined (with hub genes being fixed)
ncorr   = sum(sum(direct));
map     = (pv <= (p_discover/ncorr));
cids    = find(sum(map,1)); % non-zero columns, i.e. modulator identified
rids    = find(sum(map,2)); % non-zeros rows, i.e. conditional targets inferred

% after modulator has been inferred, use less stringent correction for
% finding the affected targets
% pv_corr = pv.*length(candidate_target);
% map     = (pv <= (p_discover/length(candidate_target)));
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Output results
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
result       = map.*max_score;
ms.data      = result(rids,cids);
ms.pv        = pv(rids,cids);
clear map result pv;

% get target annotation and correlation with the hub
target.affyid      = candidate_target(rids);
[C,IA,IB]          = intersect(target.affyid,gep.affyid);
[IC,I]             = sort(IA);
idx                = IB(I);
target.symbol      = gep.symbol(idx);
target.gene        = gep.gene(idx);
hid                = strmatch(hub,gep.affyid,'exact');
% target.correlation = corr(gep.data(hid,:)', gep.data(idx,:)')';
foo = zeros(length(idx),1,'single');
for i=1:length(idx)
    temp   = corrcoef(gep.data(hid,:),gep.data(idx(i),:));
    foo(i) = temp(1,2);
end
target.correlation = foo;

% get modulator annotation and mutual information with the hub
modulator.affyid     = candidate_modulator(cids);
[C,IA,IB]            = intersect(modulator.affyid,gep.affyid);
[IC,I]               = sort(IA);
modulator.symbol     = gep.symbol(IB(I));
modulator.gene       = gep.gene(IB(I));
[Z IX IY]            = intersect(modulator.affyid,edges);
[IZ I]               = sort(IX);
modulator.mutualinfo = mi(IY(I));

ms.target    = target;
ms.modulator = modulator;

info.hub            = hub;
info.ncandidate     = length(candidate_modulator);
info.neighborhood   = length(candidate_target);
info.percent        = percent;
info.p_independ     = p_independ;
info.p_dynamics     = p_dynamics;
info.p_discover     = p_discover;
info.p_mode         = p_mode;
info.correction     = ncorr;
ms.info             = info;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%   Determining the mode of modulation
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
data      = gep.data;
idx       = find(data<=0);
data(idx) = NaN;
mode      = zeros(size(ms.data),'single');
for i=1:size(ms.data,2) % iterate through each modulator
    idx = find(ms.data(:,i)~=0);
    if length(idx)==0
        error('Fatal error!\n');
    end
    mid = strmatch(ms.modulator.affyid{i},gep.affyid,'exact');
    for j=1:length(idx)
        rho   = ms.target.correlation(idx(j)); 
        tid   = strmatch(ms.target.affyid{idx(j)},gep.affyid,'exact');
        if isempty(tid) | length(tid)>1
            error(['Indexing error: ' ms.target.affyid{idx(j)}]);
        end
        [f,p] = determineModulatorType(log(data(mid,:)),log(data(tid,:)),p_mode);
        if p<=p_mode
            if sign(f*rho)>0
                mode(idx(j),i) = 1;
            else
                mode(idx(j),i) = -1;
            end
        end
    end
end
ms.mode = mode;
clear gep data mode;
save([outdir 'mindy_' hub '.mat'],'ms');

% compressing result and deleting the original one
if exist([tempdir cam_file],'file')
    unix(['rm ' tempdir cam_file]);
end

if exist([tempdir cmm_file],'file')
    unix(['rm ' tempdir cmm_file]);
end
